import re


def min_edit_distance(word1, word2):
    """
    最小编辑距离计算
    :param word1:
    :param word2:
    :return:
    """
    m = len(word1)
    n = len(word2)
    # 初始化二维数组，dp[i][j]表示将word1的前i个字符转换为word2的前j个字符所需的最小编辑距离
    dp = [[0] * (n + 1) for _ in range(m + 1)]
    # 初始化第一行和第一列
    for i in range(m + 1):
        dp[i][0] = i
    for j in range(n + 1):
        dp[0][j] = j
    # 动态规划填表
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if word1[i - 1] == word2[j - 1]:
                dp[i][j] = dp[i - 1][j - 1]  # 相同字符，不需要操作
            else:
                dp[i][j] = min(
                    dp[i - 1][j] + 1,  # 删除操作
                    dp[i][j - 1] + 1,  # 插入操作
                    dp[i - 1][j - 1] + 1
                )  # 替换操作
    # 返回右下角的值，即word1转换为word2的最小编辑距离
    return dp[m][n]


def remove_whitespace(text: str):
    """使用正则表达式去除所有空格和不可见字符"""
    return re.sub(r'\s+', '', text)



from enum import Enum


ROOT_TITLE_LEVEL = 0
UNKOWN_TITLE_LEVEL = 100

CONVERTED_PDF_PATH_KEY = "converted_pdf_path"
JSON_HEADERS = {
    "Content-Type": "application/json; charset=UTF-8"
}

TABLE_FILL_HEADER_PREIX = "fill_header_"


class TChunk(str, Enum):
    IMAGE = "image"
    TABLE = "table"
    TEXT = "text"


class TSentence(str, Enum):
    IMAGE = "image"
    TABLE = "table"
    TEXT = "text"


class TPdf(str, Enum):
    TEXT = "text"
    SCANNED = "scanned"


class TAtomItem(str, Enum):
    TEXT = "text"
    TABLE = "table"
    IMAGE = "image"
    IMAGE_CAPTION = "image_caption"
    TITLE = "title"
    MD_REF = "md_ref"
    MD_CODE = "md_code"
    MD_LIST = "md_list"
    MD_HTML = "md_html"


class TFileExt(str, Enum):
    MD = "md"
    DOC = "doc"
    DOCX = "docx"
    TXT = "txt"
    PDF = "pdf"
    HTML = "html"
    XLS = "xls"
    XLSX = "xlsx"
    PPT = "ppt"
    PPTX = "pptx"



from typing import List

from pptx import Presentation
from pptx.shapes.group import GroupShape

from pfip.base.constant import TFileExt, TAtomItem
from pfip.base.parser.base import Parser
from pfip.base.parser.parser_model import ParseResult, TitleNode, TitleAtomItem, AtomItem
from pfip.ppt.ppt_ele_handler import PPTTableHandler, PPTNoteHandler, \
    PPTTitleHandler, PPTSlideHandler, PPTShapeHandler, PPTParagraphHandler
from pfip.ppt.ppt_result_repair import PPTRePair


class PPTParser(Parser):
    """
        支持提取PPT的标题,表格及备注信息等

        暂不支持的特性:
        1. 不处理ppt中的图片
        2. 不支持多表头处理
    """
    title_ele_handler: PPTSlideHandler = PPTTitleHandler()
    node_ele_handler: PPTSlideHandler = PPTNoteHandler()
    shape_ele_handlers: List[PPTShapeHandler] = [PPTParagraphHandler(), PPTTableHandler()]
    ppt_repair: PPTRePair = PPTRePair()
    """pdf元素处理器集合"""

    def support(self, file_ext: str) -> bool:
        return file_ext.lower() == TFileExt.PPTX

    def build_titles(self, title_items: List[TitleAtomItem]) -> List[TitleNode]:
        titles = [item.create_and_connect() for item in title_items]
        self.fill_titles(titles)
        return titles

    def process_shape(self, shape, page_num: int) -> List[AtomItem]:
        rtn = []
        for shape_handler in self.shape_ele_handlers:
            if shape_handler.support(shape):
                rtn.extend(shape_handler.run(shape, page_num))
        return rtn

    def __call__(self, file_path: str, **kwargs) -> ParseResult:
        items = [TitleAtomItem.from_file(file_path)]
        prs = Presentation(file_path)
        for idx, slide in enumerate(prs.slides):
            page_num = idx + 1
            if self.title_ele_handler.support(slide):
                items.extend(self.title_ele_handler.run(slide, page_num))
            for shape in slide.shapes:
                if isinstance(shape, GroupShape):
                    for son_shape in shape.shapes:
                        items.extend(self.process_shape(son_shape, page_num))
                else:
                    items.extend(self.process_shape(shape, page_num))
            if self.node_ele_handler.support(slide):
                items.extend(self.node_ele_handler.run(slide, page_num))
        items = self.ppt_repair.run(items)
        title_items = [item for item in items if item.item_type == TAtomItem.TITLE]
        titles = self.build_titles(title_items)
        return ParseResult(
            items=items,
            titles=titles
        )



import os
import uuid
from abc import ABC, abstractmethod
from itertools import groupby
from typing import Any, List

import pandas as pd
from docx import Document
from docx.oxml import CT_Tbl, CT_P
from docx.text.paragraph import Paragraph
from lxml.etree import XPathEvalError
from pydantic import BaseModel

from pfip.base.errors import UnImplementException
from pfip.base.parser.parser_model import AtomItem, TableAtomItem, ImageAtomItem, WordTitleAtomItem, WordTextAtomItem
from pfip.word.auto_num_helper import WordAutoNumHelper


class WordElementHandler(ABC, BaseModel):
    @abstractmethod
    def support(self, doc: Document, ele: Any) -> bool:
        raise UnImplementException()

    @abstractmethod
    def run(self, doc: Document, ele: Any, home_dir: str, auto_num_helper: WordAutoNumHelper) -> List[AtomItem]:
        raise UnImplementException()

    @staticmethod
    def try_get_auto_num(doc: Document, ele: Any, auto_num_helper: WordAutoNumHelper) -> str:
        if not isinstance(ele, CT_P):
            return ""
        paragraph = Paragraph(ele, doc.element.body)
        return auto_num_helper.get_paragraph_num(paragraph)


class WordTitleHandler(WordElementHandler):
    def support(self, doc: Document, ele: Any) -> bool:
        if isinstance(ele, CT_P):
            return ele.style and "Heading" in doc.styles[ele.style].name
        else:
            return False

    def run(self, doc: Document, ele: Any, home_dir: str, auto_num_helper: WordAutoNumHelper) -> List[
        WordTitleAtomItem]:
        style_name = doc.styles[ele.style].name
        title_name = ele.text.strip()
        if not title_name:
            return []
        else:
            auto_num_text = self.try_get_auto_num(doc, ele, auto_num_helper)
            title_name = auto_num_text + title_name
            title_level = style_name.split()[-1]
            title_item = WordTitleAtomItem.instance(title_name, title_level)
            title_item.auto_num = auto_num_text
            return [title_item]


class WordImageHandler(WordElementHandler):

    def support(self, doc: Document, ele: Any) -> bool:
        try:
            return len(ele.xpath('.//pic:pic')) > 0
        except XPathEvalError:
            return False


    def run(self, doc: Document, ele: Any, home_dir: str, auto_num_helper: WordAutoNumHelper) -> List[ImageAtomItem]:
        image_items = []
        image_ids = ele.xpath('.//a:blip/@r:embed')
        for img_id in image_ids:
            image_part = doc.part.related_parts[img_id]
            image_path = os.path.join(home_dir, str(uuid.uuid4()) + ".png")
            with open(image_path, "wb") as f:
                f.write(image_part.blob)
            image_items.append(ImageAtomItem.simple(image_url=image_path))
        return image_items


class WordTableHandler(WordElementHandler):
    def support(self, doc: Document, ele: Any) -> bool:
        return isinstance(ele, CT_Tbl)

    @staticmethod
    def adjust_table(lines: List[List[Any]]) -> List[List[Any]]:
        # 获取列的基准数量
        sorted_lines = sorted(lines, key=lambda x: len(x))
        grouped_lines = {key: list(group) for key, group in groupby(sorted_lines, key=lambda x: len(x))}
        base_col_num = max(grouped_lines, key=lambda x: len(grouped_lines[x]))
        # 多退少补
        rtn = []
        for line in lines:
            adjust_num = len(line) - base_col_num
            if adjust_num == 0:
                pass
            elif adjust_num > 0:
                for _ in range(adjust_num):
                    line.pop()
            else:
                for _ in range(abs(adjust_num)):
                    line.append("")
            rtn.append(line)
        return rtn

    def run(self, doc: Document, ele: Any, home_dir: str, auto_num_helper: WordAutoNumHelper) -> List[TableAtomItem]:
        lines = []
        for row in ele.tr_lst:
            line = []
            for c in row.tc_lst:
                cell_val = []
                for p in c.p_lst:
                    auto_num_text = self.try_get_auto_num(doc, p, auto_num_helper)
                    cell_val.append(auto_num_text + p.text)
                line.append("".join(cell_val))
            lines.append(line)
        adjust_lines = self.adjust_table(lines)
        df = pd.DataFrame(adjust_lines[1:], columns=adjust_lines[0])
        table_item = TableAtomItem.simple(df)
        return [table_item]


class WordParagraphHandler(WordElementHandler):
    def support(self, doc: Document, ele: Any) -> bool:
        return isinstance(ele, CT_P)

    def run(self, doc: Document, ele: Any, home_dir: str, auto_num_helper: WordAutoNumHelper) -> List[WordTextAtomItem]:
        if ele.text:
            auto_num_text = self.try_get_auto_num(doc, ele, auto_num_helper)
            text = auto_num_text + ele.text.strip()
            item = WordTextAtomItem.instance(text)
            item.auto_num = auto_num_text
            return [item]
        else:
            return []



from typing import List, Any

from docx import Document

from pfip.base.constant import TFileExt, TAtomItem, CONVERTED_PDF_PATH_KEY
from pfip.base.parser.base import Parser
from pfip.base.parser.parser_model import ParseResult, AtomItem, TitleNode, TitleAtomItem
from pfip.base.parser.repair import ImageDescRepair, TableDescRepair
from pfip.word.auto_num_helper import WordAutoNumHelper
from pfip.word.word_ele_handler import WordElementHandler, WordParagraphHandler, WordTitleHandler, WordImageHandler, \
    WordTableHandler
from pfip.word.word_result_repair import PageNumRepair


class WordParser(Parser):
    """
        暂不支持的内容:
            1)超链接提取
            2)表格中的选择框
    """
    ele_handlers: List[WordElementHandler] = [
        WordTitleHandler(), WordImageHandler(), WordTableHandler()
    ]
    """word元素处理器集合"""
    fallback_handler: WordParagraphHandler = WordParagraphHandler()
    """兜底的处理器"""
    page_num_repair: PageNumRepair = PageNumRepair()
    image_desc_repair: ImageDescRepair = ImageDescRepair()
    table_desc_repair: TableDescRepair = TableDescRepair()

    def support(self, file_ext: str) -> bool:
        return file_ext.lower() == TFileExt.DOCX

    def parse_one(self, doc: Document, ele: Any, home_dir: str, auto_num_helper: WordAutoNumHelper) -> List[AtomItem]:
        handler = None
        for h in self.ele_handlers:
            if h.support(doc, ele):
                handler = h
                break
        if not handler:
            handler = self.fallback_handler
        return handler.run(doc, ele, home_dir, auto_num_helper)

    def build_titles(self, items: List[AtomItem]) -> List[TitleNode]:
        titles = [item.create_and_connect() for item in items if item.item_type == TAtomItem.TITLE]
        self.fill_titles(titles)
        return titles

    def __call__(self, file_path: str, **kwargs) -> ParseResult:
        doc = Document(file_path)
        home_dir = self.get_home_dir(file_path)
        all_items = []
        auto_num_helper = WordAutoNumHelper(file_path)
        for ele in doc.element.body:
            items = self.parse_one(doc, ele, home_dir, auto_num_helper)
            all_items.extend(items)
        # repair items
        all_items = self.image_desc_repair.run(all_items)
        all_items = self.table_desc_repair.run(all_items)
        if CONVERTED_PDF_PATH_KEY in kwargs:
            converted_pdf_path = kwargs[CONVERTED_PDF_PATH_KEY]
            all_items = self.page_num_repair.run(all_items, converted_pdf_path)
        all_items.insert(0, TitleAtomItem.from_file(file_path))
        titles = self.build_titles(all_items)
        return ParseResult(
            titles=titles,
            items=all_items
        )



from typing import List, Dict

from pydantic import BaseModel

from pfip.base.constant import TAtomItem
from pfip.base.parser.parser_model import AtomItem, ImageCaptionAtomItem, TitleNode, TextAtomItem, TitleAtomItem
from pfip.pdf.title.bookmark_title import BookMarkTitleExtractor
from pfip.pdf.title.title_item_connector import TitleConnector


class ImageRepair(BaseModel):
    """
        1)补全图片的描述信息
    """

    @staticmethod
    def _find_image_desc_item(idx: int, items: List[AtomItem]) -> tuple[int, ImageCaptionAtomItem] | tuple[int, None]:
        if idx + 1 < len(items):
            right = items[idx + 1]
            if isinstance(right, ImageCaptionAtomItem):
                return idx + 1, right
        if idx - 1 >= 0:
            left = items[idx - 1]
            if isinstance(left, ImageCaptionAtomItem):
                return idx - 1, left
        return -1, None

    def run(self, items: List[AtomItem]):
        image_item_idxs = []
        need_remove_item_idxs = []
        for idx, item in enumerate(items):
            if item.item_type == TAtomItem.IMAGE:
                image_item_idxs.append(idx)
        for idx in image_item_idxs:
            image_desc_idx, image_desc_item = self._find_image_desc_item(idx, items)
            if image_desc_item:
                items[idx].content = image_desc_item.content
                need_remove_item_idxs.append(image_desc_idx)
        filtered_items = [item for idx, item in enumerate(items) if idx not in need_remove_item_idxs]
        return filtered_items


class BookMarkTitleRepair(BaseModel):
    """
    TODO 问题待修复
    3==>1.1.1 数据的定义和生命周期／1
    2==>1.1 数据工程相关概念／1
    """

    book_mark_extractor: BookMarkTitleExtractor = BookMarkTitleExtractor()
    tconn: TitleConnector

    @staticmethod
    def to_text_item(items: List[AtomItem]) -> List[AtomItem]:
        dealed_items = []
        for item in items:
            if item.item_type == TAtomItem.TITLE:
                text_item = TextAtomItem.model_validate(item.model_dump())
                text_item.item_type = TAtomItem.TEXT
                dealed_items.append(text_item)
            else:
                dealed_items.append(item)
        return dealed_items

    @staticmethod
    def get_candcates(has_index_items: List[tuple[AtomItem, int]], page_numer: int) -> List[tuple[AtomItem, int]]:
        return [item for item in has_index_items if item[0].start_page == page_numer or item[0].end_page == page_numer]


    def run(self, file_path: str, items: List[AtomItem]) -> List[AtomItem]:
        title_nodes = self.book_mark_extractor(file_path)
        if not title_nodes:
            return items
        text_items = self.to_text_item(items)
        has_index_items: List[tuple[AtomItem, int]] = []
        for index, item in enumerate(text_items):
            has_index_items.append((item, index))
        title_conn_pos_dict: Dict[int, TitleNode] = {}

        for title in title_nodes:
            candcates = self.get_candcates(has_index_items, title.page_number)
            new_pos = self.tconn.find_similar_title(title, candcates)
            if new_pos < 0:
                continue
            title_conn_pos_dict[new_pos] = title
        # 替换处理
        final_items = []
        for index, item in enumerate(text_items):
            if index in title_conn_pos_dict:
                title = title_conn_pos_dict[index]
                title_item = TitleAtomItem.instance(title.title, title.title_level, item.start_page)
                final_items.append(title_item)
            else:
                final_items.append(item)
        return final_items



import re
from typing import List, Any, Optional

from bs4 import BeautifulSoup
from pandas import DataFrame
from pydantic import BaseModel, Field, ConfigDict, computed_field

from pfip.base.constant import TAtomItem, ROOT_TITLE_LEVEL
from pfip.base.util.file_util import get_file_name_without_ext
from pfip.base.util.generator_id import id_generator


# ==================标题===============
class TitleNode(BaseModel):
    """文档标题节点"""
    id: str = Field(default_factory=lambda: str(id_generator.next_id()))
    title: str
    title_level: int
    full_title: str = None
    """ 基于上级的标题+自身的拼接而成 """
    pid: str | None = None
    page_number: int | None = None
    sn: int = 0


# ==================文档内容===============
class Location(BaseModel):
    """左侧与页面左侧的距离。"""
    x0: Optional[float] = None
    """右侧与页面左侧的距离。"""
    x1: Optional[float] = None
    """下方与页面底部的距离。"""
    y0: Optional[float] = None
    """页面顶部末端的距离。"""
    y1: Optional[float] = None


class AtomItem(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    """原子文档内容项"""
    item_type: TAtomItem
    content: Any
    start_page: int
    end_page: int
    location: Optional[Location] = None

    @computed_field
    def full_content(self) -> str:
        raise ""


class TextAtomItem(AtomItem):
    """普通的文本项"""
    item_type: TAtomItem = TAtomItem.TEXT
    content: str

    @classmethod
    def instance(cls, content: str, page_num: int = 1) -> "TextAtomItem":
        return cls(content=content, start_page=page_num, end_page=page_num)

    @computed_field
    def full_content(self) -> str:
        return self.content


class WordTextAtomItem(TextAtomItem):
    auto_num: str = Field(default="", description="自动编号")

    @computed_field
    def without_auto_num_content(self) -> str:
        """
        不包含自动页码的content
        """
        return self.content.replace(self.auto_num, "")


class ImageAtomItem(AtomItem):
    """图片"""
    item_type: TAtomItem = TAtomItem.IMAGE
    content: str | None
    image_url: str

    @classmethod
    def simple(cls, image_url: str, content: str | None = None, page_num: int = 1) -> "ImageAtomItem":
        return cls(image_url=image_url, content=content, start_page=page_num, end_page=page_num)

    @computed_field
    def full_content(self) -> str:
        desc = self.content or "没有描述"
        return f"![{desc}]({self.image_url})"


class ImageCaptionAtomItem(AtomItem):
    """图片描述信息 目前仅版面分析模型支持"""
    item_type: TAtomItem = TAtomItem.IMAGE_CAPTION
    content: str

    @classmethod
    def simple(cls, content: str, page_num: int = 1) -> "ImageCaptionAtomItem":
        return cls(content=content, start_page=page_num, end_page=page_num)

    @computed_field
    def full_content(self) -> str:
        return self.content


class TableAtomItem(AtomItem):
    """表格"""
    item_type: TAtomItem = TAtomItem.TABLE
    content: Optional[str] = Field(default=None, description="表格描述信息")
    df: DataFrame = Field(exclude=True)

    @classmethod
    def simple(cls, df: DataFrame, desc: str = None, page_num: int = 1) -> "TableAtomItem":
        return cls(df=df, content=desc, start_page=page_num, end_page=page_num)

    @computed_field
    def table_desc(self) -> str:
        header = self.df.columns.tolist()
        dealed_header = [re.sub(r'\s+', '', str(h)) for h in header if h and str(h).strip()]
        desc = '|'.join(dealed_header)
        if self.content:
            desc = f"{self.content},表头信息:{desc}"
        return desc

    @computed_field
    def full_content(self) -> str:
        html_table = self.df.to_html(justify="center")
        if self.content:
            soup = BeautifulSoup(html_table, 'html.parser')
            caption = soup.new_tag('caption')
            caption.string = self.content
            table = soup.find('table')
            table.insert(0, caption)
            html_table = table.prettify()
        return html_table


class ScannedPDFTableAtomItem(AtomItem):
    """表格"""
    item_type: TAtomItem = TAtomItem.TABLE
    content: Optional[str] = Field(default=None, description="表格描述信息")
    table_html: str

    @classmethod
    def simple(cls, table_html: str, desc: str = None, page_num: int = 1) -> "ScannedPDFTableAtomItem":
        return cls(table_html=table_html, content=desc, start_page=page_num, end_page=page_num)

    @computed_field
    def table_desc(self) -> str:
        soup = BeautifulSoup(self.table_html, 'html.parser')
        rows = soup.find_all(["th", 'tr'])
        header = [data.text for data in rows[0].find_all('td')]
        dealed_header = [re.sub(r'\s+', '', h) for h in header if h and str(h).strip()]
        desc = '|'.join(dealed_header)
        if self.content:
            desc = f"{self.content},表头信息:{desc}"
        return desc

    @computed_field
    def full_content(self) -> str:
        html_table = self.table_html
        if self.content:
            soup = BeautifulSoup(html_table, 'html.parser')
            caption = soup.new_tag('caption')
            caption.string = self.content
            table = soup.find('table')
            table.insert(0, caption)
            html_table = table.prettify()
        return html_table


class TextPDFTableAtomItem(TableAtomItem):
    """表格"""
    page_height: float = Field(default=0.0, exclude=True)
    replace_items: List[TextAtomItem] = Field(default=[], exclude=True)
    """对应的重复文本item"""

    @classmethod
    def simple(cls, df: DataFrame, page_height: float, desc: str = None, page_num: int = 1) -> "TextPDFTableAtomItem":
        return cls(content=desc, df=df, page_height=page_height, start_page=page_num, end_page=page_num)


class TitleAtomItem(TextAtomItem):
    """标题"""
    item_type: TAtomItem = TAtomItem.TITLE
    content: str
    title_level: int
    title_node: Optional[TitleNode] = Field(default=None, exclude=True)

    @classmethod
    def instance(cls, content: str, title_level: int, page_num: int = 1) -> "TitleAtomItem":
        return cls(content=content, title_level=title_level, start_page=page_num, end_page=page_num)

    @classmethod
    def from_file(cls, file_path: str, mkpos: bool = False) -> "TitleAtomItem":
        title = get_file_name_without_ext(file_path)
        item = cls(
            content=title,
            title_level=ROOT_TITLE_LEVEL,
            start_page=1,
            end_page=1
        )
        if mkpos:
            location = Location(x0=0.0, x1=0.0, y0=0.0, y1=0.0)
            item.location = location
        return item

    def create_and_connect(self) -> TitleNode:
        """
        基于TitleAtomItem生成标题节点并与自身关联
        """
        tn = TitleNode(
            title=self.content,
            page_number=self.start_page,
            title_level=self.title_level
        )
        self.title_node = tn
        return tn


class WordTitleAtomItem(TitleAtomItem):
    auto_num: str = Field(default="", description="自动编号")

    @computed_field
    def without_auto_num_content(self) -> str:
        """
        不包含自动页码的content
        """
        return self.content.replace(self.auto_num, "")


class MDListAtomItem(AtomItem):
    item_type: TAtomItem = TAtomItem.MD_LIST
    content: List[str]

    @classmethod
    def simple(cls, content: List[str]) -> "MDListAtomItem":
        return cls(content=content, start_page=1, end_page=1)

    @computed_field
    def full_content(self) -> str:
        return "\n".join(self.content)


class MDCodeAtomItem(AtomItem):
    item_type: TAtomItem = TAtomItem.MD_CODE
    content: List[str]
    language: str

    @classmethod
    def simple(cls, content: List[str], language: str) -> "MDCodeAtomItem":
        return cls(content=content, language=language, start_page=1, end_page=1)

    @computed_field
    def full_content(self) -> str:
        return "\n".join(self.content)


class MDRefAtomItem(AtomItem):
    item_type: TAtomItem = TAtomItem.MD_REF
    content: List[str]

    @classmethod
    def simple(cls, content: List[str]) -> "MDRefAtomItem":
        return cls(content=content, start_page=1, end_page=1)

    @computed_field
    def full_content(self) -> str:
        return "\n".join(self.content)


class MDHtmlAtomItem(AtomItem):
    item_type: TAtomItem = TAtomItem.MD_HTML
    content: List[str]

    @classmethod
    def simple(cls, content: List[str]) -> "MDHtmlAtomItem":
        return cls(content=content, start_page=1, end_page=1)

    @computed_field
    def full_content(self) -> str:
        return "\n".join(self.content)


class AtomItemList(BaseModel):
    items: List[AtomItem]

    @computed_field
    def images(self) -> List[ImageAtomItem]:
        return [item for item in self.items if item.item_type == TAtomItem.IMAGE]

    @computed_field
    def titles(self) -> List[TitleAtomItem]:
        return [item for item in self.items if item.item_type == TAtomItem.TITLE]

    @computed_field
    def tables(self) -> List[TableAtomItem]:
        return [item for item in self.items if item.item_type == TAtomItem.TABLE]

    @computed_field
    def texts(self) -> List[TextAtomItem]:
        return [item for item in self.items if item.item_type == TAtomItem.TEXT]


class ParseResult(BaseModel):
    titles: List[TitleNode] = []
    items: List[AtomItem] = []

    @computed_field()
    def content(self) -> str:
        item_contents = [item.full_content for item in self.items]
        return "\n".join(item_contents)

    @computed_field()
    def content2(self) -> str:
        item_contents = ["[item_start]" + item.full_content + "[item_end]" for item in self.items]
        return "\n".join(item_contents)



import re
from abc import ABC, abstractmethod
from typing import List, Any, Tuple

from pydantic import BaseModel, Field, model_validator

from pfip.base.constant import TAtomItem
from pfip.base.parser.parser_model import TitleNode, AtomItem
from pfip.base.util.kangxi2chinese import TxtCleaner
from pfip.pdf.title.base import TitleExtractor


class PageItem(BaseModel):
    num: int
    lines: List[str]


class TocRule(ABC):
    @abstractmethod
    def is_toc(self, page: PageItem) -> bool:
        pass


class TocRule1(TocRule):
    min_scale: float = 0.7

    def is_toc(self, page: PageItem) -> bool:
        if len(page.lines) == 0:
            return False
        toc_line_num = 0
        pattern = re.compile(r'^[0-9].*')
        for l in page.lines:
            if re.match(pattern, l):
                toc_line_num += 1
        scale = toc_line_num / len(page.lines)
        return scale >= self.min_scale


class TocPageJudge:
    """
    管理Rule,判定输入页码是否为目录页
    """

    def __init__(self):
        self.rules: List[TocRule] = [TocRule1()]

    def judge(self, page: PageItem) -> bool:
        for rule in self.rules:
            if rule.is_toc(page):
                return True
        return False


class DetectTocPagesHandler(BaseModel):
    """
        用于检测目录所在页码
    """
    max_check_page_num: int

    def remove_chars(self, s: str) -> str:
        # 移除每个字符串中的所有非数字、汉字和英文字母
        s0 = TxtCleaner().aswhole(s)
        pattern = re.compile(r'[^\w\d\u4e00-\u9fa5]')
        s1 = pattern.sub('', s0)
        s2 = s1.removeprefix("第")
        s3 = self.replace_chinese_number_prefix(s2)
        s4 = self.roman_to_int(s3)
        return s4

    @staticmethod
    def roman_to_int(input_str):
        roman_dict = {'I': "1", 'V': "5", 'X': "10"}

        # 使用正则表达式匹配字符串中的罗马数字部分
        pattern = re.compile(r'^[IVX]+', re.IGNORECASE)
        match = pattern.match(input_str)
        if match:
            roman_numeral = match.group()
            for char in roman_numeral:
                if char.upper() in roman_dict:
                    input_str = input_str.replace(char, roman_dict[char.upper()], 1)  # 将汉字数字替换为阿拉伯数字，只替换一次
        return input_str

    def page_deal(self, ps: List[PageItem]) -> List[PageItem]:
        dealed = []
        for p in ps:
            dl = []
            for l in p.lines:
                line = self.remove_chars(l)
                if len(line) > 0:
                    dl.append(self.remove_chars(l))
            dealed.append(PageItem(num=p.num, lines=dl))
        return dealed

    @staticmethod
    def group_consecutive_numbers(nums):
        """
        获取最大的连贯分组
        :param nums:
        :return:
        """
        nums.sort()
        groups = []
        current_group = [nums[0]]

        for i in range(1, len(nums)):
            if nums[i] - nums[i - 1] == 1:
                current_group.append(nums[i])
            else:
                groups.append(current_group)
                current_group = [nums[i]]

        groups.append(current_group)

        max_group = max(groups, key=len)

        return max_group

    @staticmethod
    def replace_chinese_number_prefix(input_str):
        chinese_nums = {'一': '1', '二': '2', '三': '3', '四': '4', '五': '5', '六': '6', '七': '7', '八': '8',
                        '九': '9',
                        '十': '10'}
        pattern = "^[" + "".join(chinese_nums.keys()) + "]+"

        match = re.match(pattern, input_str)
        if match:
            chinese_prefix = match.group()
            for char in chinese_prefix:
                if char in chinese_nums:
                    input_str = input_str.replace(char, chinese_nums[char], 1)  # 将汉字数字替换为阿拉伯数字，只替换一次
        return input_str

    def load(self, text_items: List[AtomItem]) -> List[PageItem]:
        _pages = []
        max_page_num = min(self.max_check_page_num, text_items[-1].start_page)
        for page_idx in range(max_page_num):
            page_num = page_idx + 1
            lines = [item.content for item in text_items if item.start_page == page_num]
            _pages.append(PageItem(num=page_num, lines=lines))
        return _pages

    def __call__(self, text_items: List[AtomItem]) -> List[int]:
        pages = self.load(text_items)
        # 预处理
        dealed_pages = self.page_deal(pages)
        page_judge = TocPageJudge()
        toc_pages = []
        for dp in dealed_pages:
            if page_judge.judge(dp):
                toc_pages.append(dp)

        if len(toc_pages) == 0:
            return []
        # 确定最终的目录页
        num_list = [item.num for item in toc_pages]
        g1 = self.group_consecutive_numbers(num_list)
        final_pages = [item.num for item in toc_pages if item.num in g1]
        return final_pages


class TocPageTitleExtractor(TitleExtractor):
    handler: DetectTocPagesHandler = Field(exclude=True)

    @model_validator(mode='before')
    def init(cls, data: Any) -> Any:
        if "max_check_page_num" in data and isinstance(data["max_check_page_num"], int):
            max_check_page_num = data["max_check_page_num"]
        else:
            max_check_page_num = 10
        data["handler"] = DetectTocPagesHandler(max_check_page_num=max_check_page_num)
        return data

    @staticmethod
    def _load(items: List[AtomItem], toc_page_numbers: List[int]):
        result = []
        for item in items:
            if item.start_page in toc_page_numbers:
                result.append(item)
        return result

    @staticmethod
    def is_filter(item: AtomItem) -> bool:
        if item.content.upper() in ["目录", "TABLEOFCONTENTS", "CONTENTS"]:
            return True
        else:
            return False

    def extract_title_content(self, catalog_title: str) -> str:
        title = None
        for extract_rule in self.extract_rules:
            data = extract_rule.extract(catalog_title)
            if data:
                title = data[0]
                break
        return title

    @staticmethod
    def define_level_type_by_indent(title_with_indents: List[Tuple[str, float]]) -> List[Tuple[str, int]]:
        """根据缩进距离,确定级别大小"""
        if title_with_indents:
            indents = [ti[1] for ti in title_with_indents]
            unique_indents = list(set(indents))
            sorted_unique_indents = sorted(unique_indents)
            differences = [j - i for i, j in zip(sorted_unique_indents[:-1], sorted_unique_indents[1:])]
            # 初始化分组列表
            groups = []
            group = [sorted_unique_indents[0]]
            # 根据差值分组
            for i, diff in enumerate(differences):
                if diff > 5:
                    groups.append(group)
                    group = [sorted_unique_indents[i + 1]]
                else:
                    group.append(sorted_unique_indents[i + 1])
            groups.append(group)
            result_dict = {}
            for idx, group in enumerate(groups):
                for element in group:
                    result_dict[element] = idx + 1
            return [(ti[0], result_dict[ti[1]]) for ti in title_with_indents]
        else:
            return []

    @staticmethod
    def deal_titles(title_with_indents: List[Tuple[str, float]]) -> List[Tuple[str, float]]:
        """
            处理TITLE
            1. 去除异常title
            2. 提取title信息
        """
        rtn = []
        # 去除异常title
        need_remove_idx = []
        error_title_pattern1 = r'^[\d.]+$'
        for idx, ti in enumerate(title_with_indents):
            if re.match(error_title_pattern1, ti[0]):
                need_remove_idx.append(idx)
        # 提取title信息
        for idx, ti in enumerate(title_with_indents):
            if idx in need_remove_idx:
                continue
            # 省略号格式: '法律声明.............I'
            match1 = re.search(r'(.*?)\.{3,}(.*)', ti[0])
            if match1:
                rtn.append((match1.groups()[0], ti[1]))
                continue
            rtn.append(ti)
        return rtn

    def __call__(self, pdf_path: str, items: List[AtomItem]) -> List[TitleNode]:
        """
        1) 确定级别
        2) 上下级关系构建
        3) page_number 不需要
        """
        text_items = [item for item in items if item.item_type == TAtomItem.TEXT]
        if not text_items:
            return []
        toc_page_numbers = self.handler(text_items)
        if not toc_page_numbers:
            return []
        multi_page_items = self._load(text_items, toc_page_numbers)
        title_with_indents = []
        for item in multi_page_items:
            if self.is_filter(item):
                continue
            title_with_indents.append((item.content, item.location.x0))
        dealed_title_with_indents = self.deal_titles(title_with_indents)
        title_with_lv = self.define_level_type_by_indent(dealed_title_with_indents)
        title_nodes = []
        for idx, tlv in enumerate(title_with_lv):
            td = TitleNode(
                title=re.sub(r'\s+', '', tlv[0]),
                title_level=tlv[1],
                sn=idx + 1
            )
            title_nodes.append(td)
        return title_nodes



import re

from docx import Document
from docx.oxml.ns import qn
from docx.text.paragraph import Paragraph


class WordAutoNumHelper:
    ideographTraditional = "甲乙丙丁戊己庚辛壬癸"
    ideographZodiac = "子丑寅卯辰巳午未申酉戌亥"


    def __init__(self, file_path, gap_text="\t"):
        self.docx = Document(file_path)
        self.numId2style = self.get_style_data()
        self.gap_text = gap_text
        self.cnt = {}
        self.cache = {}
        self.result = []

    def get_paragraph_num(self, paragraph: Paragraph):
        if not self.numId2style:
            return ""
        flag1 = hasattr(paragraph, "_element")
        flag2 = hasattr(paragraph._element, "pPr")
        flag3 = hasattr(paragraph._element.pPr, "numPr")
        if flag1 and flag2 and flag3:
            return self.get_number_text(paragraph._element.pPr.numPr)
        else:
            return ""

    def get_style_data(self):
        try:
            numbering_part = self.docx.part.numbering_part._element
            abstractId2numId = {num.abstractNumId.val: num.numId for num in numbering_part.num_lst}
            numId2style = {}
            for abstractNumIdTag in numbering_part.findall(qn("w:abstractNum")):
                abstractNumId = abstractNumIdTag.get(qn("w:abstractNumId"))
                numId = abstractId2numId[int(abstractNumId)]
                for lvlTag in abstractNumIdTag.findall(qn("w:lvl")):
                    ilvl = lvlTag.get(qn("w:ilvl"))
                    style = {tag.tag[tag.tag.rfind("}") + 1:]: tag.get(qn("w:val"))
                             for tag in lvlTag.xpath("./*[@w:val]", namespaces=numbering_part.nsmap)}
                    if "numFmt" not in style:
                        numFmtVal = lvlTag.xpath(
                            "./mc:AlternateContent/mc:Fallback/w:numFmt/@w:val",
                            namespaces=numbering_part.nsmap
                        )
                        if numFmtVal and numFmtVal[0] == "decimal":
                            numFmt_format = lvlTag.xpath(
                                "./mc:AlternateContent/mc:Choice/w:numFmt/@w:format",
                                namespaces=numbering_part.nsmap
                            )
                            if numFmt_format:
                                style["numFmt"] = "decimal" + numFmt_format[0].split(",")[0]
                    if style.get("numFmt") == "decimalZero":
                        style["numFmt"] = "decimal01"
                    numId2style[(numId, int(ilvl))] = style
            return numId2style
        except Exception:
            return None


    @staticmethod
    def int2upperLetter(num):
        result = []
        while num > 0:
            num -= 1
            remainder = num % 26
            result.append(chr(remainder + ord('A')))
            num //= 26
        return "".join(reversed(result))

    @staticmethod
    def int2upperRoman(num):
        t = [
            (1000, 'M'), (900, 'CM'), (500, 'D'),
            (400, 'CD'), (100, 'C'), (90, 'XC'),
            (50, 'L'), (40, 'XL'), (10, 'X'),
            (9, 'IX'), (5, 'V'), (4, 'IV'), (1, 'I')
        ]
        roman_num = ''
        i = 0
        while num > 0:
            val, syb = t[i]
            for _ in range(num // val):
                roman_num += syb
                num -= val
            i += 1
        return roman_num

    @staticmethod
    def int2cardinalText(num):
        if not isinstance(num, int) or num < 0 or num > 999999999:
            raise ValueError(
                "Invalid number: must be a positive integer within four digits"
            )
        base = ["Zero", "One", "Two", "Three", "Four", "Five", "Six",
                "Seven", "Eight", "Nine", "Ten", "Eleven", "Twelve", "Thirteen", "Fourteen",
                "Fifteen", "Sixteen", "Seventeen", "Eighteen", "Nineteen"]
        tens = ["", "", "Twenty", "Thirty", "Fourty",
                "Fifty", "Sixty", "Seventy", "Eighty", "Ninety"]
        thousands = ["", "Thousand", "Million", "Billion"]

        def two_digits(n):
            if n < 20:
                return base[n]
            ten, unit = divmod(n, 10)
            if unit == 0:
                return f"{tens[ten]}"
            else:
                return f"{tens[ten]}-{base[unit]}"

        def three_digits(n):
            hundred, rest = divmod(n, 100)
            if hundred == 0:
                return two_digits(rest)
            result = f"{base[hundred]} hundred "
            if rest > 0:
                result += two_digits(rest)
            return result.strip()

        if num < 99:
            return two_digits(num)
        chunks = []
        while num > 0:
            num, remainder = divmod(num, 1000)
            chunks.append(remainder)
        words = []
        for i in range(len(chunks) - 1, -1, -1):
            if chunks[i] == 0:
                continue
            chunk_word = three_digits(chunks[i])
            if thousands[i]:
                chunk_word += f" {thousands[i]}"
            words.append(chunk_word)
        words = " ".join(words).lower()
        return words[0].upper() + words[1:]

    def int2ordinalText(self, num):
        if not isinstance(num, int) or num < 0 or num > 999999:
            raise ValueError(
                "Invalid number: must be a positive integer within four digits"
            )
        base = ["Zero", "One", "Two", "Three", "Four", "Five", "Six",
                "Seven", "Eight", "Nine", "Ten", "Eleven", "Twelve", "Thirteen", "Fourteen",
                "Fifteen", "Sixteen", "Seventeen", "Eighteen", "Nineteen"]
        baseth = ['Zeroth', 'First', 'Second', 'Third', 'Fourth', 'Fifth', 'Sixth', 'Seventh',
                  'Eighth', 'Ninth', 'Tenth', 'Eleventh', 'Twelfth', 'Thirteenth', 'Fourteenth',
                  'Fifteenth', 'Sixteenth', 'Seventeenth', 'Eighteenth', 'Nineteenth', 'Twentieth']
        tens = ["", "", "Twenty", "Thirty", "Fourty",
                "Fifty", "Sixty", "Seventy", "Eighty", "Ninety"]
        tensth = ["", "", "Twentieth", "Thirtieth", "Fortieth",
                  "Fiftieth", "Sixtieth", "Seventieth", "Eightieth", "Ninetieth"]

        def two_digits(n):
            if n <= 20:
                return baseth[n]
            ten, unit = divmod(n, 10)
            result = tensth[ten]
            if unit != 0:
                result = f"{tens[ten]}-{baseth[unit]}"
            return result

        thousand, num = divmod(num, 1000)
        result = []
        if thousand > 0:
            if num == 0:
                return f"{self.int2cardinalText(thousand)} thousandth"
            result.append(f"{self.int2cardinalText(thousand)} thousand")
        hundred, num = divmod(num, 100)
        if hundred > 0:
            if num == 0:
                result.append(f"{base[hundred]} hundredth")
                return " ".join(result)
            result.append(f"{base[hundred]} hundred")
        result.append(two_digits(num))
        result = " ".join(result).lower()
        return result[0].upper() + result[1:]

    @staticmethod
    def int2Chinese(num, ch_num, units):
        if not (0 <= num <= 99999999):
            raise ValueError("仅支持小于一亿以内的正整数")

        def int2Chinese_in(num, ch_num, units):
            if not (0 <= num <= 9999):
                raise ValueError("仅支持小于一万以内的正整数")
            result = [ch_num[int(i)] + unit for i, unit in zip(reversed(str(num).zfill(4)), units)]
            result = "".join(reversed(result))
            zero_char = ch_num[0]
            result = re.sub(f"(?:{zero_char}[{units}])+", zero_char, result)
            result = result.rstrip(units[0])
            if result != zero_char:
                result = result.rstrip(zero_char)
            if result.lstrip(zero_char).startswith("一十"):
                result = result.replace("一", "")
            return result

        if num < 10000:
            result = int2Chinese_in(num, ch_num, units)
        else:
            left = num // 10000
            right = num % 10000
            result = int2Chinese_in(left, ch_num, units) + "万" + int2Chinese_in(right, ch_num, units)
        if result != ch_num[0]:
            result = result.strip(ch_num[0])
        return result

    def int2ChineseCounting(self, num):
        return self.int2Chinese(num, ch_num='〇一二三四五六七八九', units='个十百千')

    def int2ChineseLegalSimplified(self, num):
        return self.int2Chinese(num, ch_num='零壹贰叁肆伍陆柒捌玖', units='个拾佰仟')

    def get_number_text(self, numpr):
        try:
            if numpr is None or numpr.numId.val == 0:
                return ""
            numId = numpr.numId.val
            ilvl = numpr.ilvl.val
            style = self.numId2style[(numId, ilvl)]
            numFmt: str = style.get("numFmt")
            lvlText = style.get("lvlText")
            if (numId, ilvl) in self.cnt:
                self.cnt[(numId, ilvl)] += 1
            else:
                self.cnt[(numId, ilvl)] = int(style["start"])
            pos = self.cnt[(numId, ilvl)]
            num_text = str(pos)
            if numFmt.startswith('decimal'):
                num_text = num_text.zfill(numFmt.count("0") + 1)
            elif numFmt == 'upperRoman':
                num_text = self.int2upperRoman(pos)
            elif numFmt == 'lowerRoman':
                num_text = self.int2upperRoman(pos).lower()
            elif numFmt == 'upperLetter':
                num_text = self.int2upperLetter(pos)
            elif numFmt == 'lowerLetter':
                num_text = self.int2upperLetter(pos).lower()
            elif numFmt == 'ordinal':
                num_text = f"{pos}{'th' if 11 <= pos <= 13 else {1: 'st', 2: 'nd', 3: 'rd'}.get(pos % 10, 'th')}"
            elif numFmt == 'cardinalText':
                num_text = self.int2cardinalText(pos)
            elif numFmt == 'ordinalText':
                num_text = self.int2ordinalText(pos)
            elif numFmt == 'ideographTraditional':
                if 1 <= pos <= 10:
                    num_text = self.ideographTraditional[pos - 1]
            elif numFmt == 'ideographZodiac':
                if 1 <= pos <= 12:
                    num_text = self.ideographZodiac[pos - 1]
            elif numFmt == 'chineseCounting':
                num_text = self.int2ChineseCounting(pos)
            elif numFmt == 'chineseLegalSimplified':
                num_text = self.int2ChineseLegalSimplified(pos)
            elif numFmt == 'decimalEnclosedCircleChinese':
                pass
            self.cache[(numId, ilvl)] = num_text
            for i in range(0, ilvl + 1):
                lvlText = lvlText.replace(f'%{i + 1}', self.cache.get((numId, i), ""))
            suff_text = {"space": " ", "nothing": ""}.get(style.get("suff"), self.gap_text)
            lvlText += suff_text
            return lvlText
        except Exception:
            return ""



import os
import re
import tempfile

import urllib.parse


def is_local_path(path):
    return path.startswith('/')


def is_http_path(path):
    return urllib.parse.urlparse(path).scheme in ('http', 'https')


def get_temp_dir() -> str:
    temp_dir = tempfile.gettempdir()
    return os.path.join(temp_dir, "pfip")


def get_file_name(file_path: str) -> str:
    """
    eg:
        in :  "./data/智能问答和智能搜索的区别有哪些？.txt"
        out: 智能问答和智能搜索的区别有哪些？.txt
    """
    file_name_with_ext = os.path.basename(file_path)
    return file_name_with_ext


def get_file_name_without_ext(file_path: str):
    """
    eg:
        in :  "./data/智能问答和智能搜索的区别有哪些？.txt"
        out: 智能问答和智能搜索的区别有哪些？
    """
    file_name_with_ext = os.path.basename(file_path)
    file_name, file_ext = os.path.splitext(file_name_with_ext)
    return file_name


def get_file_ext(file_path: str):
    file_name, file_ext = os.path.splitext(file_path)
    file_ext = file_ext[1:]
    return file_ext


def clean_file_name(file_name):
    # 使用正则表达式去除特殊符号，只保留数字、中文、英文和下划线
    cleaned_name = re.sub(r'[^\w\u4e00-\u9fa5]', '', file_name)  # \w 匹配字母、数字和下划线，\u4e00-\u9fa5 匹配中文字符
    return cleaned_name.strip()


from typing import List

from pydantic import BaseModel

from pfip.base.constant import TPdf
from pfip.base.util.rectangle import RectangleHelper
from pfip.base.util.str_util import clean_str


class PdfAnaLyserResult(BaseModel):
    pdf_type: TPdf
    """pdf类型"""
    valid_page_start_index: int = 0
    """有效页面起始索引"""
    odd_header_y1: float | None = None
    """奇数页页眉最下方的y轴坐标"""
    even_header_y1: float | None = None
    header_list: List[str] = []
    """偶数页页眉最下方的y轴坐标"""
    odd_footer_y0: float | None = None
    """奇数页页脚最上方的y轴坐标"""
    even_footer_y0: float | None = None
    """偶数页页脚最上方的y轴坐标"""
    footer_list: List[str] = []


class PdfLineBox(BaseModel):
    bottom_left: tuple[float, float]
    top_right: tuple[float, float]
    content: str


class PdfRectBox(BaseModel):
    line_boxs: List[PdfLineBox]

    @classmethod
    def from_pdf_dict(cls, rect_dict, y0: float) -> "PdfRectBox":
        line_boxs = []
        for block in rect_dict["blocks"]:
            if block['type'] == 0:
                for line in block["lines"]:
                    line_bbox = line["bbox"]
                    line_text = ""
                    for span in line["spans"]:
                        line_text += span["text"]
                    line_box = PdfLineBox(
                        bottom_left=(line_bbox[0], line_bbox[1] + y0),
                        top_right=(line_bbox[2], line_bbox[3] + y0),
                        content=clean_str(line_text)
                    )

                    line_boxs.append(line_box)
        return cls(line_boxs=line_boxs)


def get_rect_content(boxs: List[PdfRectBox], rect_box: tuple[float, float, float, float]) -> List[str]:
    """
        boxs: 一个box代表一个页面
        从box获取包含在矩形rect_box的所有文本内容
    """
    rect_angle_helper = RectangleHelper(
        bottom_left=(rect_box[0], rect_box[1]),
        top_right=(rect_box[2], rect_box[3]),
    )
    page_contents = []
    for box in boxs:
        content = ""
        for line_box in box.line_boxs:
            flag = rect_angle_helper.intersection(
                bottom_left=line_box.bottom_left,
                top_right=line_box.top_right
            )
            if flag:
                content += line_box.content
        page_contents.append(content)
    return page_contents


def get_analyse_pages(total_page_num: int, max_analyse_page_num: int) -> tuple[
    List[int], List[int]]:
    """获取待分析的页码: 奇数页码与偶数页码"""
    odd_pages = []
    even_pages = []
    # 跳过前面可能不规范的页码
    start_page_num = max(min(total_page_num - max_analyse_page_num, 10), 0)
    end_page_num = min(total_page_num, start_page_num + max_analyse_page_num)
    for page_num in range(start_page_num, end_page_num):
        if page_num % 2 != 0:
            odd_pages.append(page_num)
        else:
            even_pages.append(page_num)
    return odd_pages, even_pages



from typing import List

from pydantic import BaseModel, Field, computed_field

from pfip.base.constant import TSentence, TChunk
from pfip.base.parser.parser_model import TitleNode, AtomItem


class Sentence(BaseModel):
    item_type: TSentence
    content: str


class TextSentence(Sentence):
    item_type: TSentence = TSentence.TEXT


class ImageSentence(Sentence):
    item_type: TSentence = TSentence.IMAGE
    image_url: str


class TableSentence(Sentence):
    item_type: TSentence = TSentence.TABLE
    table_html: str


class Chunk(BaseModel):
    content: str
    title: TitleNode
    items: List[AtomItem] = Field(exclude=True)
    sentences: List[Sentence] = []
    chunk_type: TChunk
    start_page: int
    end_page: int

    @computed_field
    def title_id(self) -> str:
        return self.title.id

class SplitResult(BaseModel):
    titles: List[TitleNode]
    chunks: List[Chunk]



from typing import List

import pandas as pd

from pfip.base.constant import TFileExt, TAtomItem
from pfip.base.parser.base import Parser
from pfip.base.parser.parser_model import ParseResult, TableAtomItem, TitleNode, TitleAtomItem, AtomItem


class ExcelParser(Parser):
    """
        将sheet视为titleNode
        将sheet中的内容视为一个TableAtomItem
        暂不支持的特性:
        1. 不处理excel中的图片
        2. 不支持处理复杂表头
    """

    def support(self, file_ext: str) -> bool:
        return file_ext.lower() in [TFileExt.XLS, TFileExt.XLSX]

    def build_titles(self, items: List[AtomItem]) -> List[TitleNode]:
        title_items = [item for item in items if item.item_type == TAtomItem.TITLE]
        titles = [ti.create_and_connect() for ti in title_items]
        self.fill_titles(titles)
        return titles

    @staticmethod
    def build_items(file_path: str) -> List[AtomItem]:
        items = [TitleAtomItem.from_file(file_path)]
        excel = pd.ExcelFile(file_path)
        sheet_names = excel.sheet_names
        for idx, sheet_name in enumerate(sheet_names):
            all_df = excel.parse(sheet_name)
            df_cleaned = all_df.dropna(how='all')
            if not df_cleaned.empty:
                items.append(TitleAtomItem.instance(sheet_name, 1, idx + 1))
                items.append(TableAtomItem.simple(df_cleaned, page_num=idx + 1))
        return items


    def __call__(self, file_path: str, **kwargs) -> ParseResult:
        items = self.build_items(file_path)
        titles = self.build_titles(items)
        return ParseResult(
            items=items,
            titles=titles
        )



from typing import List, Optional, Literal

from pydantic import BaseModel, ConfigDict, Field, model_validator
from typing_extensions import Self

from pfip.base.errors import UnSupportedFileType
from pfip.base.parser.base import Parser
from pfip.base.parser.parser_model import ParseResult, AtomItem, TitleNode
from pfip.base.splitter.base import ChunkSplitter, SentenceSplitter
from pfip.base.splitter.recursive_splitter import RecursiveChunkSplitter
from pfip.base.splitter.sentence_splitter import CommonSentenceSplitter, PdfSentenceSplitter, MdSentenceSplitter, \
    NoItemSentenceSplitter
from pfip.base.splitter.splitter_model import SplitResult, Chunk
from pfip.base.splitter.title_splitter import TitleChunkSplitter, PdfTitleChunkSplitter
from pfip.base.util.file_util import get_file_ext
from pfip.excel.excel_parser import ExcelParser
from pfip.md.md_parser import MDParser
from pfip.pdf.proxy_pdf_parser import ProxyPdfParser
from pfip.ppt.ppt_parser import PPTParser
from pfip.txt.txt_parser import TxtParser
from pfip.web.html_parser import HtmlParser
from pfip.word.word_parser import WordParser


class ProcessResult(BaseModel):
    content: str = Field(description="全文内容")
    items: List[AtomItem]
    titles: List[TitleNode]
    chunks: List[Chunk]


class ParserFactory(BaseModel):
    embedding_server_url: Optional[str] = Field(default=None)
    paddle_server_url: Optional[str] = Field(default=None)
    model_config = ConfigDict(arbitrary_types_allowed=True)
    parsers: List[Parser] = Field(default=[], exclude=True)

    @model_validator(mode="after")
    def init(self) -> Self:
        self.parsers.append(ExcelParser())
        self.parsers.append(HtmlParser())
        self.parsers.append(MDParser())
        self.parsers.append(PPTParser())
        self.parsers.append(TxtParser())
        self.parsers.append(WordParser())
        self.parsers.append(
            ProxyPdfParser(
                embedding_server_url=self.embedding_server_url,
                paddle_server_url=self.paddle_server_url
            )
        )
        return self

    def get(self, file_path: str) -> Parser:
        ext = get_file_ext(file_path)
        for parser in self.parsers:
            if parser.support(ext):
                return parser
        raise UnSupportedFileType(f"不支持的文件类型:{ext}")


class ChunkSplitterFactory(BaseModel):
    splitters: List[ChunkSplitter] = Field(default=[], exclude=True)

    @model_validator(mode="after")
    def init(self) -> Self:
        self.splitters.append(PdfTitleChunkSplitter())
        self.splitters.append(TitleChunkSplitter())
        return self

    def get(self, file_path: str, spilt_type: Literal["title", "recursive"], **kwargs) -> ChunkSplitter:
        if spilt_type == "recursive":
            return RecursiveChunkSplitter(**kwargs)
        else:
            ext = get_file_ext(file_path)
            for splitter in self.splitters:
                if splitter.support(ext):
                    return splitter
            raise UnSupportedFileType(f"不支持的文件类型:{ext}")


class SentenceSplitterFactory(BaseModel):
    splitters: List[SentenceSplitter] = Field(default=[], exclude=True)

    @model_validator(mode="after")
    def init(self) -> Self:
        self.splitters.append(MdSentenceSplitter())
        self.splitters.append(PdfSentenceSplitter())
        self.splitters.append(CommonSentenceSplitter())

    def get(self, file_path: str, spilt_type: Literal["title", "recursive"]) -> SentenceSplitter:
        if spilt_type == "recursive":
            return NoItemSentenceSplitter()
        else:
            ext = get_file_ext(file_path)
            for splitter in self.splitters:
                if splitter.support(ext):
                    return splitter
            raise UnSupportedFileType(f"不支持的文件类型:{ext}")


class PanshiFileIntelligentProcess(BaseModel):
    embedding_server_url: Optional[str] = Field(default=None)
    paddle_server_url: Optional[str] = Field(default=None)
    parser_factory: ParserFactory = Field(default=None, exclude=True)
    chunk_splitter_factory: ChunkSplitterFactory = Field(default=None, exclude=True)
    sentence_splitter_factory: SentenceSplitterFactory = Field(default=None, exclude=True)

    @model_validator(mode="after")
    def init(self) -> Self:
        self.parser_factory = ParserFactory(
            embedding_server_url=self.embedding_server_url,
            paddle_server_url=self.paddle_server_url
        )
        self.chunk_splitter_factory = ChunkSplitterFactory()
        self.sentence_splitter_factory = SentenceSplitterFactory()
        return self

    def parse(self, file_path: str, **kwargs) -> ParseResult:
        """
            只解析
            word:
                converted_pdf_path : 当文档是word时,需要传递word对应的pdf文件,用于修复页码信息
        """
        parser = self.parser_factory.get(file_path)
        return parser(file_path, **kwargs)

    def split(self, file_path: str, result: ParseResult, spilt_type: Literal["title", "recursive"], **kwargs
    ) -> SplitResult:
        chunk_splitter = self.chunk_splitter_factory.get(file_path, spilt_type, **kwargs)
        sentence_splitter = self.sentence_splitter_factory.get(file_path, spilt_type)
        chunks: List[Chunk] = chunk_splitter(result)
        for chunk in chunks:
            sentences = sentence_splitter(chunk)
            chunk.sentences = sentences
        return SplitResult(
            titles=result.titles,
            chunks=chunks
        )

    def process(self, file_path: str, spilt_type: Literal["title", "recursive"] = "title", **kwargs) -> ProcessResult:
        """
        word:
            converted_pdf_path : 当文档是word时,需要传递word对应的pdf文件,用于修复页码信息

        if spilt_type= recursive,**kwargs可以设置以下参数：
            keep_separator：是否保留分隔符在生成的文本块中。
            strip_whitespace：是否移除文本块首尾的空白字符
            separators： 用于分割文本的分隔符优先级列表。
            chunk_size：单个文本块的最大字符数
            chunk_overlap：相邻文本块的重叠字符数

        """
        parse_result = self.parse(file_path)
        split_result = self.split(file_path, parse_result, spilt_type, **kwargs)
        return ProcessResult(
            content=parse_result.content,
            items=parse_result.items,
            titles=split_result.titles,
            chunks=split_result.chunks
        )



from typing import Union, Literal, List

from pydantic import model_validator, Field
from typing_extensions import Self

from pfip.base.constant import TChunk
from pfip.base.parser.parser_model import ParseResult, AtomItem
from pfip.base.splitter.base import ChunkSplitter
from pfip.base.splitter.splitter_model import Chunk
from pfip.base.util.character_splitter import RecursiveCharacterTextSplitter


class RecursiveChunkSplitter(ChunkSplitter):
    character_text_splitter: RecursiveCharacterTextSplitter = Field(default=None, exclude=True)
    separators: list[str] = ["\n\n", "\n", "。", "．", ".", "；", ";", "，", ",", " ", ""]
    keep_separator: Union[bool, Literal["start", "end"]] = False,
    is_separator_regex: bool = False,
    strip_whitespace: bool = True,
    chunk_size: int = 500,
    chunk_overlap: int = 100

    @model_validator(mode="after")
    def init(self) -> Self:
        if not self.character_text_splitter:
            self.separators.append("[item_start]")
            self.separators.append("[item_end]")
            self.character_text_splitter = RecursiveCharacterTextSplitter(
                separators=self.separators,
                keep_separator=self.keep_separator,
                is_separator_regex=self.is_separator_regex,
                strip_whitespace=self.strip_whitespace,
                chunk_size=self.chunk_size,
                chunk_overlap=self.chunk_overlap
            )

        return self

    def support(self, file_ext: str) -> bool:
        return True

    @staticmethod
    def find_item_areas(split_content: str, items: List[AtomItem], init_index: int) -> tuple[int, int]:
        """
            确定split_content涉及的起始AtomItem的范围
        """
        split_content_100_start = split_content[:100]
        split_content_100_end = split_content[-100:] if len(split_content) >= 100 else split_content

        start_index = init_index
        end_index = init_index
        # 查找起始项：匹配开头的100个字符
        for idx, item in enumerate(items):
            if idx < start_index:
                continue
            if split_content_100_start in item.full_content or item.full_content in split_content_100_start:
                start_index = idx
                break

        # 查找结束项：匹配结尾的100个字符
        for idx, item in enumerate(items):
            if idx < start_index:
                continue
            if split_content_100_end in item.full_content or item.full_content in split_content_100_end:
                end_index = idx
                break
        if start_index > end_index:
            return start_index, start_index
        return start_index, end_index


    def __call__(self, parse_result: ParseResult) -> List[Chunk]:
        split_contents = self.character_text_splitter.split_text(parse_result.content)
        chunks: List[Chunk] = []
        for split_content in split_contents:
            chunk = Chunk(
                content=split_content,
                title=parse_result.titles[0],
                items=[],
                chunk_type=TChunk.TEXT,
                start_page=1,
                end_page=1,
            )
            chunks.append(chunk)
        return chunks



from typing import List, Dict

from pydantic import Field, model_validator
from typing_extensions import Self

from pfip.base.constant import TAtomItem, TFileExt, TSentence
from pfip.base.parser.parser_model import TableAtomItem, ImageAtomItem, MDCodeAtomItem, MDListAtomItem, \
    MDRefAtomItem, TextAtomItem, AtomItem, MDHtmlAtomItem
from pfip.base.splitter.base import SentenceSplitter
from pfip.base.splitter.splitter_model import Chunk, Sentence, TableSentence, ImageSentence, TextSentence
from pfip.base.util.character_splitter import RecursiveCharacterTextSplitter, Language


class NoItemSentenceSplitter(SentenceSplitter):
    def support(self, file_ext: str) -> bool:
        return True

    def __call__(self, chunk: Chunk) -> List[Sentence]:
        return [Sentence(item_type=TSentence.TEXT, content=chunk.content)]


class CommonSentenceSplitter(SentenceSplitter):
    character_text_splitter: RecursiveCharacterTextSplitter = Field(default=None, exclude=True)
    separators: list[str] = ["。", "．", ".", "；", "；", ";", "，", ",", " ", ""]

    @model_validator(mode="after")
    def init(self) -> Self:
        if not self.character_text_splitter:
            self.character_text_splitter = RecursiveCharacterTextSplitter(
                chunk_size=self.max_sentence_size,
                chunk_overlap=0,
                separators=self.separators
            )
        return self

    def support(self, file_ext: str) -> bool:
        return True

    def custom_deal(self, item: AtomItem) -> List[Sentence]:
        return []

    def __call__(self, chunk: Chunk) -> List[Sentence]:
        sentences = []

        for item in chunk.items:
            if isinstance(item, TextAtomItem):
                sentences.extend(self.text2sentence(item))
            elif isinstance(item, TableAtomItem):
                sentences.append(self.table2sentence(item))
            elif isinstance(item, ImageAtomItem):
                sentences.append(self.image2sentence(item))
            else:
                sentences.extend(self.custom_deal(item))
        return sentences

    def text2sentence(self, item: TextAtomItem) -> List[TextSentence]:
        texts = []
        if len(item.content) > self.max_sentence_size:
            texts.extend(self.character_text_splitter.split_text(item.content))
        else:
            texts.append(item.content)
        return [TextSentence(content=text) for text in texts]

    @staticmethod
    def table2sentence(item: TableAtomItem) -> TableSentence:
        return TableSentence(content=item.table_desc, table_html=item.full_content)

    @staticmethod
    def image2sentence(item: ImageAtomItem) -> ImageSentence:
        return ImageSentence(content=item.full_content, image_url=item.image_url)


class MdSentenceSplitter(CommonSentenceSplitter):
    code_splitters: Dict[Language, RecursiveCharacterTextSplitter] = {}

    @model_validator(mode="after")
    def init_code_splitters(self) -> Self:
        for language in Language:
            language_splitter = RecursiveCharacterTextSplitter.from_language(
                language.HTML,
                chunk_size=self.max_sentence_size,
                chunk_overlap=0
            )
            self.code_splitters[language] = language_splitter
        return self

    def support(self, file_ext: str) -> bool:
        return file_ext == TFileExt.MD

    def custom_deal(self, item: AtomItem) -> List[Sentence]:
        sentences = []
        if isinstance(item, MDListAtomItem):
            sentences.extend(self.mdlist2sentence(item))
        elif isinstance(item, MDCodeAtomItem):
            sentences.extend(self.mdcode2sentence(item))
        elif isinstance(item, MDRefAtomItem):
            sentences.extend(self.mdref2sentence(item))
        elif isinstance(item, MDHtmlAtomItem):
            sentences.extend(self.mdhtml2sentence(item))
        return sentences

    def mdlist2sentence(self, item: MDListAtomItem) -> List[TextSentence]:
        texts = []
        for line in item.content:
            if len(line) > self.max_sentence_size:
                texts.extend(self.character_text_splitter.split_text(line))
            else:
                texts.append(line)
        return [TextSentence(content=text) for text in texts]

    def mdcode2sentence(self, item: MDCodeAtomItem) -> List[TextSentence]:
        texts = []
        content = "\n".join(item.content)
        if len(content) > self.max_sentence_size:
            language = item.language
            if language in self.code_splitters:
                language_splitter = self.code_splitters[language]
                texts.extend(language_splitter.split_text(content))
            else:
                texts.extend(self.character_text_splitter.split_text(content))
        else:
            texts.append(content)
        return [TextSentence(content=text) for text in texts]

    def mdref2sentence(self, item: MDRefAtomItem) -> List[TextSentence]:
        texts = []
        content = "\n".join(item.content)
        if len(content) > self.max_sentence_size:
            texts.extend(self.character_text_splitter.split_text(content))
        else:
            texts.append(content)
        return [TextSentence(content=text) for text in texts]

    def mdhtml2sentence(self, item: MDHtmlAtomItem) -> List[TextSentence]:
        texts = []
        content = "\n".join(item.content)
        if len(content) > self.max_sentence_size:
            texts.extend(self.character_text_splitter.split_text(content))
        else:
            texts.append(content)
        return [TextSentence(content=text) for text in texts]


# noinspection PyTypeChecker
class PdfSentenceSplitter(CommonSentenceSplitter):

    def support(self, file_ext: str) -> bool:
        return file_ext == TFileExt.PDF

    def __call__(self, chunk: Chunk) -> List[Sentence]:
        rtn = []
        group_items = self.group_by_image_and_table(chunk.items)
        for group in group_items:
            if isinstance(group, list):
                rtn.extend(self.items2sentence(group))
            elif isinstance(group, TableAtomItem):
                rtn.append(self.table2sentence(group))
            elif isinstance(group, ImageAtomItem):
                rtn.append(self.image2sentence(group))
            else:
                continue
        return rtn

    @staticmethod
    def group_by_image_and_table(items: List[AtomItem]) -> List[List[AtomItem] | AtomItem]:
        group_items = []
        cache_items = []
        for item in items:
            if item.item_type in [TAtomItem.TABLE, TAtomItem.IMAGE]:
                group_items.append(item)
                if cache_items:
                    group_items.append(cache_items.copy())
                    cache_items = []
            else:
                cache_items.append(item)
        if cache_items:
            group_items.append(cache_items.copy())
        return group_items

    def items2sentence(self, items: List[AtomItem]) -> List[Sentence]:
        lines = [item.content for item in items]
        all_content = "".join(lines)
        split_contents = self.character_text_splitter.split_text(all_content)
        return [TextSentence(content=text) for text in split_contents]



from typing import List, Tuple, Dict

from pydantic import model_validator, Field
from typing_extensions import Self

from pfip.base.constant import TAtomItem, TFileExt, TChunk
from pfip.base.parser.parser_model import ParseResult, AtomItem, TitleNode, TextAtomItem
from pfip.base.splitter.base import ChunkSplitter
from pfip.base.splitter.splitter_model import Chunk
from pfip.base.util.character_splitter import RecursiveCharacterTextSplitter


# noinspection PyTypeChecker
class TitleChunkSplitter(ChunkSplitter):
    """
        基于标题的段落语义切分器
    """
    character_text_splitter: RecursiveCharacterTextSplitter = Field(default=None, exclude=True)
    separators: list[str] = ["\n\n", "\n", "。", "．", ".", "；", ";", "，", ",", " ", ""]

    @model_validator(mode="after")
    def init(self) -> Self:
        if not self.character_text_splitter:
            self.character_text_splitter = RecursiveCharacterTextSplitter(
                chunk_size=self.max_chunk_size,
                chunk_overlap=0,
                separators=self.separators,
                keep_separator=True
            )
        return self

    def support(self, file_ext: str) -> bool:
        return True

    def __call__(self, parse_result: ParseResult) -> List[Chunk]:
        items_groups = self.group_by_title(parse_result.items)
        chunks = []
        for group in items_groups:
            group_chunks = self.split_group(group)
            chunks.extend(group_chunks)
        # sentence 切分
        return chunks

    @staticmethod
    def group_by_title(items: List[AtomItem]) -> List[Tuple[TitleNode, List[AtomItem]]]:
        """
            根据item_type==TAtomItem.TITLE进行分组,后面的元素为一组，直到遇见下一个TITLE元素
        """

        def add_group(_grouped_items: List, _current_group: List):
            key: str = _current_group[0].title_node
            val = _current_group[1:]
            if len(val):
                _grouped_items.append((key, val))

        grouped_items = []
        current_group = []
        for item in items:
            if item.item_type == TAtomItem.TITLE:
                if current_group:
                    add_group(grouped_items, current_group)
                    current_group = []
            current_group.append(item)
        if current_group:
            add_group(grouped_items, current_group)
        return grouped_items

    def split_group(self, group: Tuple[TitleNode, List[AtomItem]]) -> List[Chunk]:
        def create_chunk_by_cache(title: TitleNode, chunk_items: List[AtomItem]) -> Chunk:
            chunk_content = "\n".join([_item.full_content for _item in chunk_items])
            start_page = chunk_items[0].start_page
            end_page = chunk_items[-1].end_page
            return Chunk(
                title=title,
                content=chunk_content,
                items=chunk_items,
                start_page=start_page,
                end_page=end_page,
                chunk_type=TChunk.TEXT
            )

        def create_chunks_by_one(title: TitleNode, big_item: AtomItem) -> List[Chunk]:
            """
                当一个AtomItem就超过chunk最大限制时,需要再次切分
            """
            if big_item.item_type in [TAtomItem.TABLE, TAtomItem.IMAGE]:
                chunk = Chunk(
                    title=title,
                    content=big_item.full_content,
                    items=[big_item],
                    start_page=big_item.start_page,
                    end_page=big_item.end_page,
                    chunk_type=big_item.item_type
                )
                return [chunk]
            else:
                start_page = big_item.start_page
                end_page = big_item.end_page
                chunk_contents = self.character_text_splitter.split_text(big_item.full_content)
                _chunks = []
                for content in chunk_contents:
                    small_item = TextAtomItem(content=content, start_page=start_page, end_page=end_page)
                    _chunks.append(
                        Chunk(
                            title=title,
                            content=content,
                            items=[small_item],
                            start_page=start_page,
                            end_page=end_page,
                            chunk_type=TChunk.TEXT
                        )
                    )
                return _chunks

        title_node = group[0]
        items: List[AtomItem] = group[1]
        items_cache: List[AtomItem] = []
        items_cache_text_len = 0
        chunks = []
        for idx, item in enumerate(items):
            new_text_len = len(item.full_content)
            if len(item.full_content) > self.max_chunk_size \
                    or item.item_type == TAtomItem.TABLE \
                    or item.item_type == TAtomItem.IMAGE:
                """FIXED:将表格单独切成一个chunk"""
                if len(items_cache) > 0:
                    items_cache_text_len = 0
                    chunks.append(create_chunk_by_cache(title_node, items_cache.copy()))
                    items_cache.clear()
                chunks.extend(create_chunks_by_one(title_node, item))
            elif new_text_len + items_cache_text_len > self.max_chunk_size:
                # 生成chunk 清除缓存
                chunks.append(create_chunk_by_cache(title_node, items_cache.copy()))
                items_cache.clear()
                items_cache.append(item)
                items_cache_text_len = new_text_len
            else:
                items_cache.append(item)
                items_cache_text_len += new_text_len

        if items_cache_text_len > 0:
            chunks.append(create_chunk_by_cache(title_node, items_cache))
        return chunks


class PdfTitleChunkSplitter(TitleChunkSplitter):

    def support(self, file_ext: str) -> bool:
        return TFileExt.PDF.value == file_ext.lower()

    def __call__(self, parse_result: ParseResult) -> List[Chunk]:
        items_groups = self.group_by_title(parse_result.items)
        chunks = []
        for group in items_groups:
            group_chunks = self.split_group(group)
            chunks.extend(group_chunks)
        # sentence 切分
        return chunks

    def split_group(self, group: Tuple[TitleNode, List[AtomItem]]) -> List[Chunk]:
        def build_char_page_index(chunk_items: List[AtomItem]) -> Dict[int, int]:
            """
            构建 char_index=>page_num
            char_index: 字符在整个content中的索引. content是所有chunk_items的content拼接而成
            page_num:页码
            """
            char_page_dict = {}
            append_content = ""
            for _item in chunk_items:
                pre_index = len(append_content)
                assert _item.start_page == _item.end_page
                page_num = _item.start_page
                item_dict = {i: page_num for i in range(pre_index, pre_index + len(_item.full_content) + 1)}
                char_page_dict.update(item_dict)
                append_content += _item.full_content
            return char_page_dict

        def create_chunk_by_cache(title: TitleNode, chunk_items: List[AtomItem]) -> List[Chunk]:
            total_len = 0
            total_content = ""
            for _item in chunk_items:
                total_content += f"{_item.full_content}"
                total_len += len(total_content)
            if total_len > self.max_chunk_size:
                chunk_contents = self.character_text_splitter.split_text(total_content)
                char_page_index = build_char_page_index(chunk_items)
                _chunks = []
                append_part = ""
                for part in chunk_contents:
                    start_page = char_page_index[len(append_part)]
                    end_page = char_page_index[len(append_part) + len(part)]
                    _chunks.append(
                        Chunk(
                            title=title,
                            content=part,
                            items=[TextAtomItem(content=part, start_page=start_page, end_page=end_page)],
                            start_page=start_page,
                            end_page=end_page,
                            chunk_type=TChunk.TEXT
                        )
                    )
                    append_part += part
                return _chunks
            else:
                start_page = chunk_items[0].start_page
                end_page = chunk_items[-1].end_page
                return [Chunk(
                    title=title,
                    content=total_content,
                    items=chunk_items,
                    start_page=start_page,
                    end_page=end_page,
                    chunk_type=TChunk.TEXT
                )]

        def create_chunks_by_one(title: TitleNode, big_item: AtomItem) -> List[Chunk]:
            """
                当一个AtomItem就超过chunk最大限制时,需要再次切分
            """
            if big_item.item_type in [TAtomItem.TABLE, TAtomItem.IMAGE]:
                chunk = Chunk(
                    title=title,
                    content=big_item.full_content,
                    items=[big_item],
                    start_page=big_item.start_page,
                    end_page=big_item.end_page,
                    chunk_type=big_item.item_type
                )
                return [chunk]
            else:
                start_page = big_item.start_page
                end_page = big_item.end_page
                chunk_contents = self.character_text_splitter.split_text(big_item.full_content)
                _chunks = []
                for content in chunk_contents:
                    small_item = TextAtomItem(content=content, start_page=start_page, end_page=end_page)
                    _chunks.append(
                        Chunk(
                            title=title,
                            content=content,
                            items=[small_item],
                            start_page=start_page,
                            end_page=end_page,
                            chunk_type=TChunk.TEXT
                        )
                    )
                return _chunks

        title_node = group[0]
        items: List[AtomItem] = group[1]
        items_cache: List[AtomItem] = []
        chunks = []
        for idx, item in enumerate(items):
            if item.item_type == TAtomItem.TABLE or item.item_type == TAtomItem.IMAGE:
                if items_cache:
                    chunks.extend(create_chunk_by_cache(title_node, items_cache.copy()))
                    items_cache.clear()
                chunks.extend(create_chunks_by_one(title_node, item))
            else:
                items_cache.append(item)
        if items_cache:
            chunks.extend(create_chunk_by_cache(title_node, items_cache.copy()))
        return chunks



import re
from difflib import SequenceMatcher
from typing import List

import fitz
from loguru import logger
from pydantic import BaseModel

from pfip.base.util.str_util import remove_empty_lines
from pfip.pdf.analyse.base import get_analyse_pages, PdfRectBox, get_rect_content


class HeaderAnalyser(BaseModel):
    max_analyse_page_num: int = 20
    """用于页眉分析的最大页码数量"""
    start_header_percent: float = 0.2
    """起始的header位置. 从页面高度的x百分比开始"""
    reduce_step_len: int = 20
    """每次缩减的步长大小"""
    min_same_similar_threshold: float = 0.95
    """两个文本被视为相同的最小阈值"""

    def run(self, pdf_path: str) -> tuple[float, float, List[str]]:
        with fitz.open(pdf_path) as doc:
            footer_list = []
            if doc.page_count < 4:
                return 0, 0, []
            odd_pages, even_pages = get_analyse_pages(doc.page_count, self.max_analyse_page_num)
            odd_y1, odd_texts = self.find_header_area(doc, odd_pages)
            even_y1, even_texts = self.find_header_area(doc, even_pages)
            footer_list.extend(odd_texts)
            footer_list.extend(even_texts)
            return odd_y1, even_y1, list(set(footer_list))

    def find_header_area(self, doc, pages: List[int]) -> tuple[float, List[str]]:
        page_mediabox = doc[pages[0]].mediabox
        pdf_height = doc[pages[0]].rect.height
        try:
            boxs = self.extract_pages_header(doc, pages)
            header_area_y1 = pdf_height * self.start_header_percent
            while header_area_y1 > 0:
                rect = fitz.Rect(page_mediabox.x0, 0, page_mediabox.x1, header_area_y1)
                header_texts = get_rect_content(boxs, rect)
                header_texts = [re.sub(r'\d+', "[页码]", t) for t in header_texts]
                header_texts = [re.sub(r'\b[IVXLCDM]+\b', "[页码]", t) for t in header_texts]
                if self.is_header(header_texts):
                    return header_area_y1, header_texts
                header_area_y1 -= self.reduce_step_len
            return 0, []
        except Exception as e:
            logger.error(e)
            return 0, []

    def extract_pages_header(self, doc, pages: List[int]) -> tuple[float, float, List[str]]:
        """抽取指定页码,y1区域的文本信息"""
        boxs = []
        for page_num in pages:
            page = doc[page_num]
            y1 = page.rect.height * self.start_header_percent
            rect = fitz.Rect(page.mediabox.x0, 0, page.mediabox.x1, y1)
            page.set_cropbox(rect)
            box = PdfRectBox.from_pdf_dict(page.get_text("dict"), 0)
            boxs.append(box)
        return boxs

    def is_header(self, header_texts: List[str]) -> bool:
        """是否为页眉"""
        header_texts = remove_empty_lines(header_texts)
        if len(header_texts) < 2:
            return False
        middle_index = len(header_texts) // 2
        middle_element = header_texts[middle_index]
        similar_count = 0
        for ele in header_texts:
            matcher = SequenceMatcher(None, ele, middle_element)
            if matcher.ratio() > self.min_same_similar_threshold:
                similar_count += 1
        radio = (similar_count - 1) / (len(header_texts) - 1)
        if radio == 1:
            return True
        else:
            return False



import re
from abc import ABC, abstractmethod
from collections import deque
from typing import List

import pandas as pd
from pandas import DataFrame
from pydantic import BaseModel

from pfip.base.errors import UnImplementException
from pfip.base.parser.parser_model import AtomItem, TitleAtomItem, TextAtomItem, TableAtomItem, MDListAtomItem, \
    MDCodeAtomItem, MDRefAtomItem, MDHtmlAtomItem


class MDElementHandler(ABC, BaseModel):
    @abstractmethod
    def support(self, line: str) -> bool:
        raise UnImplementException()

    @abstractmethod
    def run(self, start_line: str, q: deque[str]) -> AtomItem:
        raise UnImplementException()


class MDParagraphHandler(MDElementHandler):
    def support(self, line: str) -> bool:
        return True

    def run(self, start_line: str, q: deque[str]) -> TextAtomItem:
        return TextAtomItem.instance(start_line)


class MDTitleHandler(MDElementHandler):
    reg: str = r'^(#+)\s+(.*)$'

    def support(self, line: str) -> bool:
        return bool(re.match(self.reg, line))

    def run(self, start_line: str, q: deque[str]) -> TitleAtomItem:
        match = re.match(self.reg, start_line)
        level = len(match.group(1))
        title = match.group(2)
        return TitleAtomItem.instance(title, level)


class MDTableHandler(MDElementHandler):
    reg: str = r'^\|\s*(.*?)\s*\|\s*(.*?)\s*\|$'

    def support(self, line: str) -> bool:
        return bool(re.match(self.reg, line))

    @staticmethod
    def mdtable2df(lines: List[str]) -> DataFrame:
        header = [h.strip() for h in re.split(r'\s*\|\s*', lines[0]) if h.strip()]
        data = [[d.strip() for d in re.split(r'\s*\|\s*', line) if d.strip()] for line in lines[2:]]
        return pd.DataFrame(data, columns=header)

    def run(self, start_line: str, q: deque[str]) -> TableAtomItem:
        table_lines = [start_line]
        while q:
            line = q.popleft()
            if bool(re.match(self.reg, line)):
                table_lines.append(line)
            else:
                q.appendleft(line)
                break
        df = self.mdtable2df(table_lines)
        return TableAtomItem.simple(df=df)


class MDListHandler(MDElementHandler):
    unordered_list_reg: str = r'^[\*\-\+] .*'
    ordered_list_reg: str = r'^\d+\. .*'

    def support(self, line: str) -> bool:
        return bool(re.match(self.unordered_list_reg, line)) or bool(re.match(self.ordered_list_reg, line))

    def run(self, start_line: str, q: deque[str]) -> MDListAtomItem:
        list_lines = [start_line]
        while q:
            line = q.popleft()
            if self.support(line):
                list_lines.append(line)
            else:
                q.appendleft(line)
                break
        return MDListAtomItem.simple(content=list_lines)


class MDCodeHandler(MDElementHandler):
    def support(self, line: str):
        return line.startswith("```")

    def run(self, start_line: str, q: deque[str]) -> MDCodeAtomItem:
        code_lines = [start_line]
        while q:
            line = q.popleft()
            if line == "```":
                break
            else:
                code_lines.append(line)
        language = start_line.replace("```", "")
        return MDCodeAtomItem.simple(content=code_lines, language=language)


class MDRefHandler(MDElementHandler):
    def support(self, line: str):
        return line.startswith(">")

    def run(self, start_line: str, q: deque[str]) -> MDRefAtomItem:
        ref_lines = [start_line]
        while q:
            line = q.popleft()
            if line.startswith(">"):
                ref_lines.append(line)
            else:
                q.appendleft(line)
                break
        return MDRefAtomItem.simple(content=ref_lines)


class MdHtmlHandler(MDElementHandler):
    html_start_pattern: str = r'<(\w+)\s+[^>/]*>'
    """<div xxx>"""
    html_pattern: str = r'<(\w+)\s+[^>]*>.*?</\1>'
    """ <div> xxx </div>"""
    single_line_html_pattern: str = r'<(\w+)\s+[^>]*/>'
    """ <div xxx/>"""

    def support(self, line: str):
        match1 = re.fullmatch(self.html_start_pattern, line.strip(), re.DOTALL)
        match2 = re.fullmatch(self.single_line_html_pattern, line.strip(), re.DOTALL)
        match3 = re.fullmatch(self.html_pattern, line.strip(), re.DOTALL)
        return match1 or match2 or match3


    def run(self, start_line: str, q: deque[str]) -> MDHtmlAtomItem:
        single_match = re.fullmatch(self.single_line_html_pattern, start_line.strip(), re.DOTALL)
        single_match2 = re.fullmatch(self.html_pattern, start_line.strip(), re.DOTALL)
        if single_match or single_match2:
            return MDHtmlAtomItem.simple(content=[start_line])
        else:
            html_lines = [start_line]
            while q:
                html_content = "".join(html_lines)
                match = re.fullmatch(self.html_pattern, html_content.strip(), re.DOTALL)
                if match:
                    break
                else:
                    line = q.popleft()
                    html_lines.append(line)
            return MDHtmlAtomItem.simple(content=html_lines)



from collections import deque
from typing import List, Optional

from pydantic import model_validator, Field
from typing_extensions import Self

from pfip.base.constant import TFileExt, TAtomItem
from pfip.base.parser.base import Parser
from pfip.base.parser.parser_model import ParseResult, AtomItem, TitleNode, TitleAtomItem
from pfip.md.md_ele_handler import MDElementHandler, MDParagraphHandler, MDTitleHandler, \
    MDTableHandler, MDListHandler, MDCodeHandler, MdHtmlHandler, MDRefHandler


class MDParser(Parser):
    """
        支持code|table|txt|image识别
        按行读取
        status
    """
    ele_handlers: Optional[List[MDElementHandler]] = Field(default=None, exclude=True)
    """md元素处理器集合"""
    fallback_handler: Optional[MDElementHandler] = Field(default=None, exclude=True)
    """兜底的处理器"""

    @model_validator(mode='after')
    def init_handlers(self) -> Self:
        self.fallback_handler = MDParagraphHandler()
        self.ele_handlers = [
            MDRefHandler(), MDTitleHandler(), MDTableHandler(), MDListHandler(), MDCodeHandler(), MdHtmlHandler()
        ]
        return self

    def support(self, file_ext: str) -> bool:
        return file_ext.lower() == TFileExt.MD

    def read(self, file_path: str) -> deque[str]:
        q = deque()
        with open(file_path, 'r', encoding='utf-8') as file:
            for line in file:
                if not self.is_empty_line(line):
                    q.append(line.strip())
        return q

    def find_ele_handler(self, line: str) -> MDElementHandler:
        for h in self.ele_handlers:
            if h.support(line):
                return h
        return self.fallback_handler

    def parse(self, file_path: str, q: deque[str]) -> List[AtomItem]:
        rtn = [TitleAtomItem.from_file(file_path)]
        while q:
            start_line = q.popleft()
            hander = self.find_ele_handler(start_line)
            rtn.append(hander.run(start_line, q))
        return rtn

    def build_titles(self, items: List[AtomItem]) -> List[TitleNode]:
        titles = [item.create_and_connect() for item in items if item.item_type == TAtomItem.TITLE]
        self.fill_titles(titles)
        return titles

    def __call__(self, file_path: str, **kwargs) -> ParseResult:
        q = self.read(file_path)
        items = self.parse(file_path, q)
        titles = self.build_titles(items)
        return ParseResult(
            items=items,
            titles=titles
        )



from typing import List

from pfip.base.constant import TFileExt, TAtomItem
from pfip.base.parser.base import Parser
from pfip.base.parser.parser_model import ParseResult, TitleNode, TextAtomItem, TitleAtomItem, AtomItem


class TxtParser(Parser):
    filters: List[str] = ["\n", "\r\n", " "]

    def support(self, file_ext: str) -> bool:
        return file_ext.lower() == TFileExt.TXT

    def build_item(self, file_path: str) -> List[TextAtomItem]:
        rtn = [TitleAtomItem.from_file(file_path)]
        with open(file_path, 'r', encoding='utf-8') as file:
            for line in file:
                if line in self.filters:
                    continue
                rtn.append(TextAtomItem.instance(line))
        return rtn

    def build_titles(self, items: List[AtomItem]) -> List[TitleNode]:
        titles = [item.create_and_connect() for item in items if item.item_type == TAtomItem.TITLE]
        self.fill_titles(titles)
        return titles

    def __call__(self, file_path: str, **kwargs) -> ParseResult:
        items = self.build_item(file_path)
        titles = self.build_titles(items)
        return ParseResult(
            titles=titles,
            items=items
        )



import os
from abc import ABC, abstractmethod
from typing import List

import fitz
from loguru import logger
from pandas import DataFrame
from pydantic import BaseModel
from pymupdf import Page

from pfip.base.errors import UnImplementException
from pfip.base.parser.parser_model import AtomItem, ImageAtomItem, TextAtomItem, TextPDFTableAtomItem, Location
from pfip.base.util.str_util import clean_str_keep_spaces


class TextPdfElementHandler(ABC, BaseModel):
    @abstractmethod
    def run(self, doc, page_num: int, page: Page, home_dir: str) -> List[AtomItem]:
        raise UnImplementException()


class TextPdfParagraphHandler(TextPdfElementHandler):

    def run(self, doc, page_num: int, page: Page, home_dir: str) -> List[TextAtomItem]:
        rtn = []
        text_dict = page.get_text("dict")
        for block in text_dict["blocks"]:
            if "lines" in block:
                block_text = ""
                block_bbox = block["bbox"]
                for line in block["lines"]:
                    for span in line["spans"]:
                        block_text += span["text"]
                block_text = clean_str_keep_spaces(block_text)
                if block_text:
                    item = TextAtomItem.instance(block_text, page_num)
                    item.location = Location(x0=block_bbox[0], y0=block_bbox[1], x1=block_bbox[2], y1=block_bbox[3])
                    rtn.append(item)
        return rtn


class TextPdfTableHandler(TextPdfElementHandler):
    COL_MIN_NUM: int = 2
    """表格最小列数量"""

    def _is_regular_table(self, df: DataFrame) -> bool:
        """是否是正确的表格.库普遍存在解决错误的情况"""
        # 最小列数判断
        if len(df.columns.tolist()) < self.COL_MIN_NUM:
            return False
        return True

    def run(self, doc, page_num: int, page: Page, home_dir: str) -> List[TextPDFTableAtomItem]:
        rtn = []
        tables = page.find_tables()
        height = page.rect.height
        for table_index, table in enumerate(tables):
            df = table.to_pandas()
            if not self._is_regular_table(df):
                continue
            table_item = TextPDFTableAtomItem(
                df=df,
                page_height=height,
                start_page=page_num,
                end_page=page_num
            )
            bbox = table.bbox
            table_item.location = Location(x0=bbox[0], y0=bbox[1], x1=bbox[2], y1=bbox[3])
            rtn.append(table_item)
        return rtn


class TextPdfImageHandler(TextPdfElementHandler):
    MIN_AREA: int = 50 * 100

    def _is_invalid_image(self, image_area: float, page_area: float) -> bool:
        """
            is_small 去除非常小的图片,比如说logo等
            is_big 去除非常大的图片
        """
        is_small = self.MIN_AREA > image_area
        is_big = image_area / page_area > 0.9
        return is_small or is_big

    def run(self, doc, page_num: int, page: Page, home_dir: str) -> List[ImageAtomItem]:
        rtn = []
        img_list = page.get_images(full=True)
        page_area = page.rect.height * page.rect.width
        for image_index, image in enumerate(img_list):
            try:
                xref = image[0]
                img_info = doc.extract_image(xref)
                if not page.get_image_rects(xref):
                    continue
                img_rect = page.get_image_rects(xref)[0]
                width = img_rect[2] - img_rect[0]
                height = img_rect[3] - img_rect[1]
                flag = self._is_invalid_image(width * height, page_area)
                if not img_info or flag:
                    continue
                image_path = os.path.join(home_dir, f"page_{page_num}_image_{image_index + 1}.png")
                pix = fitz.Pixmap(doc, xref)
                # 检查色彩空间并转换
                if pix.colorspace is None:
                    pix = fitz.Pixmap(fitz.csRGB, pix)  # 处理无色彩空间的图像
                elif pix.colorspace.name not in ("DeviceRGB", "DeviceGray"):
                    try:
                        pix = fitz.Pixmap(fitz.csRGB, pix)  # 尝试转换为RGB
                    except Exception as e:
                        pix = None
                if pix is not None:
                    pix.save(image_path)
                item = ImageAtomItem(
                    image_url=image_path,
                    content=None,
                    start_page=page_num,
                    end_page=page_num
                )
                item.location = Location(x0=img_rect[0], y0=img_rect[1], x1=img_rect[2], y1=img_rect[3])
                rtn.append(item)
            except Exception as e:
                logger.error(e)
        return rtn



import re

from pfip.base.util.kangxi2chinese import TxtCleaner

TXT_CLEANER = TxtCleaner()


def remove_whitespace(text: str, except_newline: bool = False):
    """使用正则表达式去除所有空格和不可见字符"""
    if except_newline:
        return re.sub(r'[^\S\n]+', '', text)
    else:
        return re.sub(r'\s+', '', text)


def clean_str(content: str, except_newline: bool = False) -> str:
    if content:
        content = TXT_CLEANER.aswhole(content)
        content = remove_whitespace(content, except_newline)
        return content
    else:
        return content


def clean_str_keep_spaces(content: str) -> str:
    if content:
        content = TXT_CLEANER.aswhole(content)
        content = re.sub(r'[\t\n\r]', '', content)
        return content
    else:
        return content


def remove_empty_lines(lines: list):
    """移除列表中的空行和只包含空白字符的行"""
    return list(filter(lambda line: line.strip(), lines))



import re
from difflib import SequenceMatcher
from typing import List

import fitz
from loguru import logger
from pydantic import BaseModel

from pfip.base.util.str_util import remove_empty_lines
from pfip.pdf.analyse.base import get_analyse_pages, PdfRectBox, get_rect_content


class FooterAnalyser(BaseModel):
    max_analyse_page_num: int = 20
    """用于页脚分析的最大页码数量"""
    start_footer_percent: float = 0.8
    """起始的footer位置. 从页面高度的x百分比开始"""
    add_step_len: int = 20
    """每次增加的步长大小"""
    min_same_similar_threshold: float = 0.9
    """两个文本被视为相同的最小阈值"""

    def run(self, pdf_path: str) -> tuple[float, float, List[str]]:
        with fitz.open(pdf_path) as doc:
            footer_list = []
            if doc.page_count < 4:
                pdf_height = doc[0].rect.height
                return pdf_height, pdf_height, []
            odd_pages, even_pages = get_analyse_pages(doc.page_count, self.max_analyse_page_num)
            odd_y0, odd_texts = self.find_footer_area(doc, odd_pages)
            even_y0, even_texts = self.find_footer_area(doc, even_pages)
            footer_list.extend(odd_texts)
            footer_list.extend(even_texts)
            return odd_y0, even_y0, list(set(footer_list))

    def find_footer_area(self, doc, pages: List[int]) -> tuple[float, List[str]]:
        page_mediabox = doc[pages[0]].mediabox
        pdf_height = doc[pages[0]].rect.height
        try:
            boxs = self.extract_pages_footer(doc, pages)
            footer_area_y0 = pdf_height * self.start_footer_percent
            while footer_area_y0 < pdf_height - self.add_step_len:
                rect = fitz.Rect(page_mediabox.x0, footer_area_y0, page_mediabox.x1, pdf_height)
                footer_texts = get_rect_content(boxs, rect)
                footer_texts = [re.sub(r'\d+', "[页码]", t) for t in footer_texts]
                footer_texts = [re.sub(r'\b[IVXLCDM]+\b', "[页码]", t) for t in footer_texts]

                if self.is_footer(footer_texts):
                    return footer_area_y0, footer_texts
                footer_area_y0 += self.add_step_len
            return pdf_height, []
        except Exception as e:
            logger.error(e)
            return pdf_height, []

    def extract_pages_footer(self, doc, pages: List[int]) -> List[PdfRectBox]:
        """抽取指定页码,y0区域的文本信息"""
        boxs = []
        for page_num in pages:
            page = doc[page_num]
            y0 = page.rect.height * self.start_footer_percent
            rect = fitz.Rect(page.mediabox.x0, y0, page.mediabox.x1, page.rect.height)
            page.set_cropbox(rect)
            box = PdfRectBox.from_pdf_dict(page.get_text("dict"), y0)
            boxs.append(box)
        return boxs

    def is_footer(self, footer_texts: List[str]) -> bool:
        """是否为页脚"""
        footer_texts = remove_empty_lines(footer_texts)
        if len(footer_texts) < 2:
            return False
        middle_index = len(footer_texts) // 2
        middle_element = footer_texts[middle_index]
        similar_count = 0
        for ele in footer_texts:
            matcher = SequenceMatcher(None, ele, middle_element)
            if matcher.ratio() > self.min_same_similar_threshold:
                similar_count += 1
        radio = (similar_count - 1) / (len(footer_texts) - 1)
        if radio == 1:
            return True
        else:
            return False



from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Optional

from panshi2task.paddle_client import PaddleTaskClient
from pydantic import Field, model_validator
from tqdm import tqdm
from typing_extensions import Self

from pfip.base.constant import TFileExt, TAtomItem
from pfip.base.parser.base import Parser
from pfip.base.parser.parser_model import ParseResult, TitleNode, AtomItem, TitleAtomItem
from pfip.pdf.pdf_splitter import PdfSplitter
from pfip.pdf.scanned.scanned_pdf_ele_handler import ScannedPdfElementHandler
from pfip.pdf.scanned.scanned_pdf_result_repair import ImageRepair, BookMarkTitleRepair
from pfip.pdf.title.title_item_connector import TitleConnector


class ScannedPdfParser(Parser):
    paddle_server_url: str
    """paddle服务地址"""
    embedding_server_url: Optional[str] = None
    thread_pool_num: int = 5
    client: PaddleTaskClient = Field(default=None, exclude=True)
    handler: ScannedPdfElementHandler = Field(default=ScannedPdfElementHandler(), exclude=True)
    image_repair: ImageRepair = Field(default=ImageRepair(), exclude=True)
    title_repair: BookMarkTitleRepair = Field(default=None, exclude=True)
    pdf_splitter: PdfSplitter = PdfSplitter(max_split_pdf_num=5)
    thread_pool: ThreadPoolExecutor = None  # 初始化线程池

    @model_validator(mode='after')
    def init(self) -> Self:
        self.client = PaddleTaskClient(self.paddle_server_url)
        connect = TitleConnector(embedding_server_url=self.embedding_server_url)
        self.title_repair = BookMarkTitleRepair(tconn=connect)
        self.thread_pool = ThreadPoolExecutor(max_workers=self.thread_pool_num)
        return self

    def __del__(self):
        if hasattr(self, 'thread_pool') and self.thread_pool:
            self.thread_pool.shutdown()


    def support(self, file_ext: str) -> bool:
        return file_ext == TFileExt.PDF

    def build_titles(self, items: List[AtomItem]) -> List[TitleNode]:
        titles: List[TitleNode] = [item.create_and_connect() for item in items if item.item_type == TAtomItem.TITLE]
        self.fill_titles(titles)
        return titles

    def process_one(self, file_path: str) -> tuple[int, List[AtomItem]]:
        _itr = self.client.pdf_structure_bytes(file_path)
        file_start_page, _ = self.pdf_splitter.get_page_area_from_name(file_path)
        items = self.handler.run(_itr, file_path)
        for item in items:
            item.start_page = item.start_page + file_start_page
            item.end_page = item.end_page + file_start_page
        return file_start_page, items

    def __call__(self, file_path: str, **kwargs) -> ParseResult:
        home_dir = self.get_home_dir(file_path)
        split_pdf_paths = self.pdf_splitter.split(file_path, home_dir)
        futures = [self.thread_pool.submit(self.process_one, path) for path in split_pdf_paths]
        results = []
        for future in tqdm(as_completed(futures), total=len(futures)):
            result = future.result()
            results.append(result)
        sorted_results = sorted(results, key=lambda x: x[0])
        items = []
        for sorted_result in sorted_results:
            items.extend(sorted_result[1])
        items = self.image_repair.run(items)
        items = self.title_repair.run(file_path, items)
        items.insert(0, TitleAtomItem.from_file(file_path, mkpos=True))
        titles = self.build_titles(items)
        self.pdf_splitter.clean(home_dir)
        return ParseResult(
            items=items,
            titles=titles
        )



from pydantic import BaseModel

from pfip.base.constant import TPdf
from pfip.pdf.analyse.base import PdfAnaLyserResult
from pfip.pdf.analyse.footer_analyser import FooterAnalyser
from pfip.pdf.analyse.header_analyser import HeaderAnalyser
from pfip.pdf.analyse.pdf_type_analyser import PdfTypeAnalyser
from pfip.pdf.analyse.valid_page_analyser import ValidPageAnalyser


class PdfAnalyser(BaseModel):
    """pdf分析器"""
    valid_page_analyser: ValidPageAnalyser = ValidPageAnalyser()
    pdf_type_analyser: PdfTypeAnalyser = PdfTypeAnalyser()
    header_analyser: HeaderAnalyser = HeaderAnalyser()
    footer_analyser: FooterAnalyser = FooterAnalyser()

    def analyse_text_pdf(self, pdf_path: str) -> PdfAnaLyserResult:
        valid_page_start_index = self.valid_page_analyser.run(pdf_path)
        odd_header_y1, even_header_y1, header_list = self.header_analyser.run(pdf_path)
        odd_footer_y0, even_footer_y0, footer_list = self.footer_analyser.run(pdf_path)
        return PdfAnaLyserResult(
            pdf_type=TPdf.TEXT,
            valid_page_start_index=valid_page_start_index,
            odd_header_y1=odd_header_y1,
            even_header_y1=even_header_y1,
            header_list=header_list,
            odd_footer_y0=odd_footer_y0,
            even_footer_y0=even_footer_y0,
            footer_list=footer_list
        )

    def __call__(self, pdf_path: str) -> PdfAnaLyserResult:
        pdf_type = self.pdf_type_analyser.run(pdf_path)
        if pdf_type == TPdf.SCANNED:
            return PdfAnaLyserResult(pdf_type=TPdf.SCANNED)
        else:
            return self.analyse_text_pdf(pdf_path)



from abc import ABC, abstractmethod
from typing import List

from pydantic import BaseModel, ConfigDict

from pfip.base.errors import UnImplementException
from pfip.base.parser.parser_model import TitleNode, AtomItem


class TitleExtractor(BaseModel, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def __call__(self, file_path: str, items: List[AtomItem]) -> List[TitleNode]:
        raise UnImplementException("未实现...")



from typing import List

import fitz
from loguru import logger
from panshi2task.paddle_client import PaddleTaskClient, BoxRecItem
from pydantic import Field, model_validator
from tqdm import tqdm
from typing_extensions import Self

from pfip.base.constant import UNKOWN_TITLE_LEVEL
from pfip.base.parser.parser_model import TitleNode, AtomItem
from pfip.base.util.file_util import get_file_name
from pfip.pdf.title.base import TitleExtractor


class LayoutAnalysisTitleExtractor(TitleExtractor):
    """
        基于版面分析模型的title构建工具.
    """
    paddle_server_url: str
    """paddle服务地址"""
    client: PaddleTaskClient = Field(default=None, exclude=True)
    max_support_pages: int = 30

    @model_validator(mode='after')
    def init(self) -> Self:
        self.client = PaddleTaskClient(self.paddle_server_url)
        return self

    @staticmethod
    def parse_title(rec_item: BoxRecItem) -> List[TitleNode]:
        rtn = []
        if not rec_item.res:
            return rtn
        title: str = ""
        for item in rec_item.res:
            if not item.text:
                continue
            title += item.text
        rtn.append(TitleNode(title=title, title_level=UNKOWN_TITLE_LEVEL))
        return rtn

    def __call__(self, file_path: str, items: List[AtomItem]=[]) -> List[TitleNode]:
        rtn = []
        with fitz.open(file_path) as pdf:
            total_page_num = pdf.page_count
        if total_page_num > self.max_support_pages:
            logger.warning("文字版PDF页码超过{}限制,耗时过长,不支持处理...", self.max_support_pages)
            return rtn
        file_name = get_file_name(file_path)
        progress_bar = tqdm(total=total_page_num, desc=f"版面分析中==>{file_name}")
        last_page_num = 0
        for rec_item in self.client.pdf_structure_bytes(file_path):
            cur_page_number = rec_item.page_num
            if rec_item.type == "title":
                rtn.extend(self.parse_title(rec_item))
            else:
                continue
            if last_page_num != cur_page_number:
                progress_bar.update(cur_page_number - last_page_num)
                last_page_num = cur_page_number
        return rtn



import os
import shutil
from typing import List

from loguru import logger
from pydantic import BaseModel
from PyPDF2 import PdfReader, PdfWriter

from pfip.base.util.file_util import get_file_name_without_ext


class PdfSplitter(BaseModel):
    max_split_pdf_num: int = 10
    split_pdf_dir: str = "splits"

    @staticmethod
    def get_split_pdf_name(start_page: int, end_page: int) -> str:
        return f"split_{start_page}_{end_page}.pdf"

    @staticmethod
    def get_page_area_from_name(file_path: str) -> tuple[int, int]:
        file_name = get_file_name_without_ext(file_path)
        if file_name.startswith("split_"):
            file_name = file_name.replace("split_", "")
            start_page = file_name.split("_")[0]
            end_page = file_name.split("_")[1]
            return int(start_page), int(end_page)
        else:
            return 0, 0

    def clean(self, home_dir: str):
        output_dir = os.path.join(home_dir, self.split_pdf_dir)
        if os.path.exists(output_dir):
            try:
                shutil.rmtree(output_dir)
            except OSError as e:
                logger.error(f"Error: {e.strerror}")

    def split(self, file_path: str, home_dir: str) -> List[str]:
        reader = PdfReader(file_path)
        total_pages = len(reader.pages)

        if total_pages < self.max_split_pdf_num:
            """不需要切分,直接返回"""
            return [file_path]

        output_dir = os.path.join(home_dir, self.split_pdf_dir)
        os.makedirs(output_dir, exist_ok=True)

        start_page = 0
        split_index = 1
        all_split_paths = []
        while start_page < total_pages:
            writer = PdfWriter()
            end_page = min(start_page + self.max_split_pdf_num, total_pages)
            split_file_name = self.get_split_pdf_name(start_page, end_page)
            for page_num in range(start_page, end_page):
                writer.add_page(reader.pages[page_num])
            output_file = os.path.join(output_dir, split_file_name)
            with open(output_file, 'wb') as output_pdf:
                writer.write(output_pdf)
            start_page = end_page
            split_index += 1
            all_split_paths.append(output_file)
        return all_split_paths



import fitz
from loguru import logger
from pydantic import BaseModel
from pymupdf import Page

from pfip.base.util.str_util import remove_empty_lines, clean_str


class ValidPageAnalyser(BaseModel):
    start_valid_page: int = 10
    """有效页校验的最大页码范围"""
    line_num_rule: int = 15
    """每页的行数不能小于10行"""
    text_len_rule: int = 200
    """每页的内容不得少于100个字符"""

    def run(self, pdf_path: str) -> int:
        with fitz.open(pdf_path) as doc:
            try:
                if doc.page_count < self.start_valid_page:
                    return 0
                last_not_valid_page_index = -1
                may_be_start_valid_page = min(self.start_valid_page, doc.page_count)
                for page_index in range(may_be_start_valid_page):
                    page = doc.load_page(page_index)
                    if self.is_valid(page):
                        break
                    else:
                        last_not_valid_page_index = page_index
                return last_not_valid_page_index + 1
            except Exception as e:
                logger.error(e)
                return 0

    def is_valid(self, page: Page) -> bool:
        text: str = page.get_text("text")
        cleaned_text = clean_str(text, except_newline=True)
        lines = remove_empty_lines(cleaned_text.split("\n"))
        line_num = len(lines)
        text_len = len(clean_str(text))
        flag1 = line_num > self.line_num_rule
        flag2 = text_len > self.text_len_rule
        return flag1 and flag2



import os
from typing import List

import html2text
from bs4 import BeautifulSoup

from pfip.base.constant import TFileExt, TAtomItem
from pfip.base.parser.base import Parser
from pfip.base.parser.parser_model import ParseResult, TextAtomItem, TitleAtomItem, AtomItem, TitleNode
from pfip.base.util.common import remove_whitespace
from pfip.base.util.file_util import get_file_name_without_ext


class HtmlParser(Parser):

    def support(self, file_ext: str) -> bool:
        return file_ext.lower() == TFileExt.HTML

    @staticmethod
    def build_item(file_path: str) -> List[TextAtomItem]:
        rtn = [TitleAtomItem.from_file(file_path)]
        with open(file_path, 'r', encoding='utf-8') as file:
            html_content = file.read()
            soup = BeautifulSoup(html_content, 'html.parser')
            text = soup.get_text()
            text_lines = [line.strip() for line in text.split('\n') if line.strip()]
            cleaned_text = '\n'.join(text_lines)
            rtn.append(TextAtomItem.instance(cleaned_text))
        return rtn

    def build_titles(self, items: List[AtomItem]) -> List[TitleNode]:
        titles = [item.create_and_connect() for item in items if item.item_type == TAtomItem.TITLE]
        self.fill_titles(titles)
        return titles

    @staticmethod
    def html2md(file_path: str, home_dir: str) -> str:
        with open(file_path, 'r', encoding='utf-8') as file:
            html_content = file.read()
        markdown_content = html2text.html2text(html_content)
        md_file_path = os.path.join(home_dir, get_file_name_without_ext(file_path) + "_convert.md")
        with open(md_file_path, 'w', encoding='utf-8') as file:
            file.write(markdown_content)
        return md_file_path

    def __call__(self, file_path: str, **kwargs) -> ParseResult:
        items = self.build_item(file_path)
        titles = self.build_titles(items)
        return ParseResult(
            titles=titles,
            items=items
        )



from typing import Optional, List, Self

import fitz
from loguru import logger
from pydantic import Field, model_validator
from pymupdf import Page, Rect, _format_g
from tqdm import tqdm

from pfip.base.constant import TFileExt, TAtomItem
from pfip.base.parser.base import Parser
from pfip.base.parser.parser_model import ParseResult, TitleAtomItem, AtomItem, TitleNode
from pfip.base.util.file_util import get_file_name
from pfip.pdf.analyse.base import PdfAnaLyserResult
from pfip.pdf.text.text_pdf_ele_handler import TextPdfElementHandler, TextPdfParagraphHandler, TextPdfTableHandler, \
    TextPdfImageHandler
from pfip.pdf.text.text_pdf_result_repair import TextPdfRepair, TextPdfTitleRepair
from pfip.pdf.title.base import TitleExtractor
from pfip.pdf.title.bookmark_title import BookMarkTitleExtractor
from pfip.pdf.title.layout_analysis_title import LayoutAnalysisTitleExtractor
from pfip.pdf.title.title_item_connector import TitleConnector
from pfip.pdf.title.toc_page_title import TocPageTitleExtractor


class TextPdfParser(Parser):
    """
     TODO : 文字版PDF掺杂着图片页的情况的处理
    """
    paddle_server_url: Optional[str] = None
    embedding_server_url: Optional[str] = None
    ele_handlers: List[TextPdfElementHandler] = Field(default=[], exclude=True)
    title_extractors: List[TitleExtractor] = Field(default=[], exclude=True)
    text_pdf_repair: TextPdfRepair = Field(default=TextPdfRepair(), exclude=True)
    title_repair: TextPdfTitleRepair = Field(default=None, exclude=True)

    @model_validator(mode='after')
    def init(self) -> Self:
        # handlers
        self.ele_handlers.append(TextPdfParagraphHandler())
        self.ele_handlers.append(TextPdfTableHandler())
        self.ele_handlers.append(TextPdfImageHandler())
        # title_extractors
        self.title_extractors.append(BookMarkTitleExtractor())
        self.title_extractors.append(TocPageTitleExtractor())
        if self.paddle_server_url:
            self.title_extractors.append(LayoutAnalysisTitleExtractor(paddle_server_url=self.paddle_server_url))
        else:
            logger.warning("paddle_server_url未设置,基于版面分析的标题提取组件将不可用...")
        if self.embedding_server_url:
            connect = TitleConnector(embedding_server_url=self.embedding_server_url)
            self.title_repair = TextPdfTitleRepair(tconn=connect)
        else:
            logger.warning("embedding_server_url未设置,标题修复组件不可用...")
        return self

    def support(self, file_ext: str) -> bool:
        return file_ext == TFileExt.PDF

    def extract_titles(self, file_path: str, items: List[AtomItem]) -> List[TitleNode]:
        rtn = []
        for title_extractor in self.title_extractors:
            nodes = title_extractor(file_path, items)
            if nodes:
                rtn.extend(nodes)
                break
        return rtn

    def build_titles(self, items: List[AtomItem]) -> List[TitleNode]:
        titles: List[TitleNode] = [item.create_and_connect() for item in items if item.item_type == TAtomItem.TITLE]
        self.fill_titles(titles)
        return titles

    @staticmethod
    def set_crop_box(doc, page_index: int, page: Page, analyse_result: PdfAnaLyserResult):
        """
            裁剪掉页眉页脚
        """
        if page_index % 2 == 0:
            # 偶数页索引(索引从0计算)
            header_y1 = analyse_result.even_header_y1
            footer_y0 = analyse_result.even_footer_y0
        else:
            header_y1 = analyse_result.odd_header_y1
            footer_y0 = analyse_result.odd_footer_y0

        rect = (
            0,
            header_y1,
            page.rect.width,
            footer_y0
        )
        try:
            mb = page.mediabox
            rect = Rect(rect[0], mb.y1 - rect[3], rect[2], mb.y1 - rect[1])
            doc.xref_set_key(page.xref, "CropBox", f"[{_format_g(tuple(rect))}]")
        except ValueError as e:
            logger.warning(e)

    def one_process(self, home_dir: str, file_path: str, analyse_result: PdfAnaLyserResult):
        rtn = []
        with fitz.open(file_path) as doc:
            file_name = get_file_name(file_path)
            for index in tqdm(range(doc.page_count), desc=file_name):
                if analyse_result:
                    if index < analyse_result.valid_page_start_index:
                        continue
                    page = doc.load_page(index)
                    self.set_crop_box(doc, index, page, analyse_result)
                else:
                    page = doc.load_page(index)
                page_num = index + 1
                for handler in self.ele_handlers:
                    rtn.extend(handler.run(doc, page_num, page, home_dir))
        return rtn

    def __call__(self, file_path: str, **kwargs) -> ParseResult:
        if kwargs.get("analyse_result"):
            analyse_result: PdfAnaLyserResult = kwargs.get("analyse_result")
        else:
            analyse_result = None
        # 解析处理
        home_dir = self.get_home_dir(file_path)
        items = self.one_process(home_dir, file_path, analyse_result)
        # 修复items
        items = self.text_pdf_repair.run(items)
        assert len(items) > 0, "空的pdf,没有可解析的元素"
        # 标题处理
        title_candcates = self.extract_titles(file_path, items)
        if self.title_repair:
            items = self.title_repair.run(title_candcates, items)
        items.insert(0, TitleAtomItem.from_file(file_path, mkpos=True))
        titles = self.build_titles(items)
        return ParseResult(
            items=items,
            titles=titles
        )



from pydantic import BaseModel, model_validator, Field, ConfigDict
from shapely import Polygon
from typing_extensions import Self


def get_rectangle_points(bottom_left, top_right):
    """
        根据交点获取逆时针坐标
    """
    x1, y1 = bottom_left
    x2, y2 = top_right
    return [(x1, y1), (x2, y1), (x2, y2), (x1, y2)]


class RectangleHelper(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    bottom_left: tuple[float, float]
    top_right: tuple[float, float]
    rectangle: Polygon = Field(default=None)

    @model_validator(mode="after")
    def init(self) -> Self:
        points = get_rectangle_points(self.bottom_left, self.top_right)
        self.rectangle = Polygon(points)

    def contains(self, bottom_left: tuple[float, float], top_right: tuple[float, float]) -> bool:
        """
            rectangle 是否包含rectangle2矩形
        """
        points = get_rectangle_points(bottom_left, top_right)
        rectangle2 = Polygon(points)
        return self.rectangle.contains(rectangle2)

    def intersection(self, bottom_left: tuple[float, float], top_right: tuple[float, float]) -> bool:
        """
            rectangle 是否与rectangle2矩形有交集
        """
        points = get_rectangle_points(bottom_left, top_right)
        rectangle2 = Polygon(points)
        if self.rectangle.intersection(rectangle2):
            return True
        else:
            return False

    def in_top(self, bottom_left: tuple[float, float], top_right: tuple[float, float]) -> bool:
        """
            PDF坐标系是反转的, center_y值越大,越在下面
            rectangle 是否在rectangle2的上面
        """
        points = get_rectangle_points(bottom_left, top_right)
        rectangle2 = Polygon(points)
        rectangle1_center = self.rectangle.centroid
        rectangle1_center_y = rectangle1_center.y
        rectangle2_center = rectangle2.centroid
        rectangle2_center_y = rectangle2_center.y
        return rectangle2_center_y > rectangle1_center_y



import re
from typing import List, Tuple

import PyPDF2
from loguru import logger
from pydantic import BaseModel

from pfip.base.constant import TAtomItem, ROOT_TITLE_LEVEL
from pfip.base.parser.parser_model import AtomItem, WordTitleAtomItem, WordTextAtomItem, TitleAtomItem, TextAtomItem


class PageNumRepair(BaseModel):
    error_page_num: int = 0

    @staticmethod
    def _remove_whitespace(text: str):
        """使用正则表达式去除所有空格和不可见字符"""
        return re.sub(r'\s+', '', text)

    def _read_pdf(self, file_path: str) -> List[str]:
        with open(file_path, 'rb') as pdf_file:
            pdf_reader = PyPDF2.PdfReader(pdf_file)
            num_pages = len(pdf_reader.pages)
            pages = []
            for page_num in range(num_pages):
                page = pdf_reader.pages[page_num]
                page_text = page.extract_text()
                pages.append(self._remove_whitespace(page_text))
            return pages

    def _compare(self, content: str, pages: List[str], cur_page_num: int) -> Tuple[int, int]:
        cur_page = pages[cur_page_num - 1]
        if content in cur_page:
            return cur_page_num, cur_page_num
        next_pages = [cur_page]
        for i in range(10):
            if cur_page_num + i >= len(pages):
                break
            next_page = pages[cur_page_num + i]
            next_pages.append(next_page)
            if content in next_page:
                return cur_page_num + i + 1, cur_page_num + i + 1
            elif content in next_pages[-2] + next_page:
                return cur_page_num + i, cur_page_num + i + 1
        return self.error_page_num, self.error_page_num

    @staticmethod
    def get_left_item(idx: int, items: List[AtomItem]) -> tuple[int, AtomItem]:
        cur_idx = idx - 1
        item = None
        while cur_idx >= 0:
            if items[cur_idx].start_page:
                item = items[cur_idx]
                break
            else:
                cur_idx -= 1
        return cur_idx, item

    @staticmethod
    def get_right_item(idx: int, items: List[AtomItem]) -> tuple[int, AtomItem]:
        cur_idx = idx + 1
        item = None
        while cur_idx < len(items):
            if items[cur_idx].start_page:
                item = items[cur_idx]
                break
            else:
                cur_idx += 1
        return cur_idx, item

    def _infer_fail_item(self, idx: int, fail_idxs: List[int], items: List[AtomItem], total_page_num: int):
        """根据上下文的页码信息进行推断"""
        fail_item = items[idx]
        lidx, left = self.get_left_item(idx, items)
        ridx, right = self.get_right_item(idx, items)
        if len(items) == idx + 1:
            right = TextAtomItem.instance("最后一页", page_num=total_page_num)
        if left is None or right is None:
            logger.error("无法修复此元素的页码信息:{}", fail_item.model_dump_json())
            return
        if ridx in fail_idxs:
            # todo 右侧元素也是需要修复的,暂时默认当前元素与前一个元素在同一页，会有误差
            fail_item.start_page = left.end_page
            fail_item.end_page = left.end_page
            return
        page_range = right.start_page - left.end_page
        if page_range > 2:
            logger.error("无法修复此元素的页码信息,相邻元素页码差距过大{},:{}", page_range, fail_item.model_dump_json())
            return
        if fail_item.item_type == TAtomItem.IMAGE:
            # 图片不能跨页
            fail_item.start_page = right.start_page
            fail_item.end_page = right.start_page
        else:
            fail_item.start_page = left.end_page
            fail_item.end_page = right.start_page

    def run(self, items: List[AtomItem], pdf_path: str):
        pages = self._read_pdf(pdf_path)
        fail_idxs = []
        cur_page_num = 1
        for idx, item in enumerate(items):
            if isinstance(item, TitleAtomItem) and item.title_level == ROOT_TITLE_LEVEL:
                continue
            if item.item_type in [TAtomItem.TABLE, TAtomItem.IMAGE]:
                fail_idxs.append(idx)
                continue
            if isinstance(item, WordTitleAtomItem) or isinstance(item, WordTextAtomItem):
                content = self._remove_whitespace(item.without_auto_num_content)
            else:
                content = self._remove_whitespace(item.content)
            start_page, end_page = self._compare(content, pages, cur_page_num)
            if start_page == self.error_page_num:
                fail_idxs.append(idx)
            else:
                item.start_page = start_page
                item.end_page = end_page
                cur_page_num = end_page
        for fail_idx in fail_idxs:
            self._infer_fail_item(fail_idx, fail_idxs, items, len(pages))
        return items



class TxtCleaner:

    def full2half_str(self, uni_str):
        half_chars = []

        for uni_char in uni_str:
            uni_code = ord(uni_char)
            if uni_code == 12288:
                uni_code = 32  # space full to half
            elif 65281 <= uni_code <= 65374:
                uni_code -= 65248
            half_char = chr(uni_code)
            half_chars.append(half_char)
        half_str = ''.join(half_chars)

        return half_str

    def full2half_alphabet_and_num(self, uni_str):
        half_chars = []

        for uni_char in uni_str:
            uni_code = ord(uni_char)
            if uni_code == 12288:
                uni_code = 32
            elif 65296 <= uni_code <= 65305 or 65313 <= uni_code <= 65338 or 65345 <= uni_code <= 65370:
                uni_code -= 65248
            half_char = chr(uni_code)
            half_chars.append(half_char)
        half_str = ''.join(half_chars)

        return half_str

    def half2full_punctuation(self, uni_str, escape_sequence='%*+-./^'):
        full_chars = []

        for uni_char in uni_str:
            uni_code = ord(uni_char)
            if (33 <= uni_code <= 47 or 58 <= uni_code <= 64 or 91 <= uni_code <= 96 or 123 <= uni_code <= 126) \
                    and uni_char not in escape_sequence:
                uni_code += 65248
            full_char = chr(uni_code)
            full_chars.append(full_char)
        full_str = ''.join(full_chars)

        return full_str

    def kangxi2hanzi(self, uni_str):
        hanzi_chars = []
        for uni_char in uni_str:
            uni_code = ord(uni_char)

            if 12032 <= uni_code <= 12255:
                if uni_code == 12032:  # 一
                    uni_code = 19968
                if uni_code == 12036:  # 乙
                    uni_code = 20057
                if uni_code == 12038:  # 二
                    uni_code = 20108
                if uni_code == 12040:  # 人
                    uni_code = 20154
                if uni_code == 12041:  # 儿
                    uni_code = 20799
                if uni_code == 12042:  # 入
                    uni_code = 20837
                if uni_code == 12043:  # 八
                    uni_code = 20843
                if uni_code == 12047:  # 几
                    uni_code = 20960
                if uni_code == 12049:  # 刀
                    uni_code = 20992
                if uni_code == 12050:  # 力
                    uni_code = 21147
                if uni_code == 12052:  # 匕
                    uni_code = 21269
                if uni_code == 12055:  # 十
                    uni_code = 21313
                if uni_code == 12056:  # 卜
                    uni_code = 21340
                if uni_code == 12058:  # 厂
                    uni_code = 21378
                if uni_code == 12060:  # 又
                    uni_code = 21448
                if uni_code == 12061:  # 口
                    uni_code = 21475
                if uni_code == 12063:  # 土
                    uni_code = 22303
                if uni_code == 12064:  # 士
                    uni_code = 22763
                if uni_code == 12067:  # 夕
                    uni_code = 22805
                if uni_code == 12068:  # 大
                    uni_code = 22823
                if uni_code == 12069:  # 女
                    uni_code = 22899
                if uni_code == 12070:  # 子
                    uni_code = 23376
                if uni_code == 12072:  # 寸
                    uni_code = 23544
                if uni_code == 12073:  # 小
                    uni_code = 23567
                if uni_code == 12075:  # 尸
                    uni_code = 23608
                if uni_code == 12077:  # 山
                    uni_code = 23665
                if uni_code == 12079:  # 工
                    uni_code = 24037
                if uni_code == 12080:  # 己
                    uni_code = 24049
                if uni_code == 12081:  # 巾
                    uni_code = 24062
                if uni_code == 12082:  # 干
                    uni_code = 24178
                if uni_code == 12083:  # 幺
                    uni_code = 24186
                if uni_code == 12084:  # 广
                    uni_code = 24191
                if uni_code == 12087:  # 弋
                    uni_code = 24331
                if uni_code == 12088:  # 弓
                    uni_code = 24339
                if uni_code == 12092:  # 心
                    uni_code = 24515
                if uni_code == 12093:  # 戈
                    uni_code = 25096
                if uni_code == 12094:  # 户
                    uni_code = 25143
                if uni_code == 12095:  # 手
                    uni_code = 25163
                if uni_code == 12096:  # 支
                    uni_code = 25903
                if uni_code == 12098:  # 文
                    uni_code = 25991
                if uni_code == 12099:  # 斗
                    uni_code = 26007
                if uni_code == 12100:  # 斤
                    uni_code = 26020
                if uni_code == 12101:  # 方
                    uni_code = 26041
                if uni_code == 12102:  # 无
                    uni_code = 26080
                if uni_code == 12103:  # 日
                    uni_code = 26085
                if uni_code == 12105:  # 月
                    uni_code = 26376
                if uni_code == 12106:  # 木
                    uni_code = 26408
                if uni_code == 12107:  # 欠
                    uni_code = 27424
                if uni_code == 12108:  # 止
                    uni_code = 27490
                if uni_code == 12109:  # 歹
                    uni_code = 27513
                if uni_code == 12111:  # 母
                    uni_code = 27597
                if uni_code == 12112:  # 比
                    uni_code = 27604
                if uni_code == 12113:  # 毛
                    uni_code = 27611
                if uni_code == 12114:  # 氏
                    uni_code = 27663
                if uni_code == 12115:  # 气
                    uni_code = 27668
                if uni_code == 12116:  # 水
                    uni_code = 27700
                if uni_code == 12117:  # 火
                    uni_code = 28779
                if uni_code == 12118:  # 爪
                    uni_code = 29226
                if uni_code == 12119:  # 父
                    uni_code = 29238
                if uni_code == 12122:  # 片
                    uni_code = 29255
                if uni_code == 12123:  # 牙
                    uni_code = 29273
                if uni_code == 12124:  # 牛
                    uni_code = 29275
                if uni_code == 12125:  # 犬
                    uni_code = 29356
                if uni_code == 12126:  # 玄
                    uni_code = 29572
                if uni_code == 12127:  # 玉
                    uni_code = 29577
                if uni_code == 12128:  # 瓜
                    uni_code = 29577
                if uni_code == 12129:  # 瓦
                    uni_code = 29926
                if uni_code == 12130:  # 甘
                    uni_code = 29976
                if uni_code == 12131:  # 生
                    uni_code = 29983
                if uni_code == 12132:  # 用
                    uni_code = 29992
                if uni_code == 12133:  # 田
                    uni_code = 30000
                if uni_code == 12137:  # 白
                    uni_code = 30333
                if uni_code == 12138:  # 皮
                    uni_code = 30382
                if uni_code == 12139:  # 皿
                    uni_code = 30399
                if uni_code == 12140:  # 目
                    uni_code = 30446
                if uni_code == 12141:  # 矛
                    uni_code = 30683
                if uni_code == 12142:  # 矢
                    uni_code = 30690
                if uni_code == 12143:  # 石
                    uni_code = 30707
                if uni_code == 12144:  # 示
                    uni_code = 31034
                if uni_code == 12146:  # 禾
                    uni_code = 31166
                if uni_code == 12147:  # 穴
                    uni_code = 31348
                if uni_code == 12148:  # 立
                    uni_code = 31435
                if uni_code == 12149:  # 竹
                    uni_code = 31481
                if uni_code == 12150:  # 米
                    uni_code = 31859
                if uni_code == 12153:  # 网
                    uni_code = 32593
                if uni_code == 12154:  # 羊
                    uni_code = 32650
                if uni_code == 12155:  # 羽
                    uni_code = 32701
                if uni_code == 12156:  # 老
                    uni_code = 32769
                if uni_code == 12157:  # 而
                    uni_code = 32780
                if uni_code == 12159:  # 耳
                    uni_code = 32819
                if uni_code == 12161:  # 肉
                    uni_code = 32905
                if uni_code == 12162:  # 臣
                    uni_code = 33251
                if uni_code == 12163:  # 自
                    uni_code = 33258
                if uni_code == 12164:  # 至
                    uni_code = 33267
                if uni_code == 12166:  # 舌
                    uni_code = 33292
                if uni_code == 12168:  # 舟
                    uni_code = 33311
                if uni_code == 12170:  # 色
                    uni_code = 33394
                if uni_code == 12173:  # 虫
                    uni_code = 34411
                if uni_code == 12174:  # 血
                    uni_code = 34880
                if uni_code == 12175:  # 行
                    uni_code = 34892
                if uni_code == 12176:  # 衣
                    uni_code = 34915
                if uni_code == 12180:  # 言
                    uni_code = 35328
                if uni_code == 12181:  # 谷
                    uni_code = 35895
                if uni_code == 12182:  # 豆
                    uni_code = 35910
                if uni_code == 12186:  # 赤
                    uni_code = 36196
                if uni_code == 12187:  # 走
                    uni_code = 36208
                if uni_code == 12188:  # 足
                    uni_code = 36275
                if uni_code == 12189:  # 身
                    uni_code = 36523
                if uni_code == 12191:  # 辛
                    uni_code = 36763
                if uni_code == 12192:  # 辰
                    uni_code = 36784
                if uni_code == 12197:  # 里
                    uni_code = 37324
                if uni_code == 12198:  # 金
                    uni_code = 37329
                if uni_code == 12204:  # 雨
                    uni_code = 38632
                if uni_code == 12206:  # 非
                    uni_code = 38750
                if uni_code == 12207:  # 面
                    uni_code = 38754
                if uni_code == 12210:  # 韭
                    uni_code = 38893
                if uni_code == 12211:  # 音
                    uni_code = 38899
                if uni_code == 12215:  # 食
                    uni_code = 39135
                if uni_code == 12216:  # 首
                    uni_code = 39318
                if uni_code == 12217:  # 香
                    uni_code = 39321
                if uni_code == 12219:  # 骨
                    uni_code = 39592
                if uni_code == 12220:  # 高
                    uni_code = 39640
                if uni_code == 12225:  # 鬼
                    uni_code = 39740
                if uni_code == 12229:  # 鹿
                    uni_code = 40575
                if uni_code == 12232:  # 黄
                    uni_code = 40644
                if uni_code == 12233:  # 黍
                    uni_code = 40653
                if uni_code == 12234:  # 黑
                    uni_code = 40657
                if uni_code == 12238:  # 鼓
                    uni_code = 40723
                if uni_code == 12239:  # 鼠
                    uni_code = 40736
                if uni_code == 12240:  # 鼻
                    uni_code = 40763
                if uni_code == 12232:  # 黄
                    uni_code = 40644

            hanzi_char = chr(uni_code)
            hanzi_chars.append(hanzi_char)
        hanzi_str = ''.join(hanzi_chars)
        return hanzi_str

    def buchong2hanzi(self, uni_str):
        hanzi_chars = []
        for uni_char in uni_str:
            uni_code = ord(uni_char)
            if 11904 <= uni_code <= 12031:
                if uni_code == 11918:  # 兀
                    uni_code = 20800
                if uni_code == 11935:  # 母
                    uni_code = 27597
                if uni_code == 11936:  # 民
                    uni_code = 27665
                if uni_code == 11969:  # 虎
                    uni_code = 34382
                if uni_code == 11972:  # 西
                    uni_code = 35199
                if uni_code == 11973:  # 见
                    uni_code = 35265
                if uni_code == 11974:  # 角
                    uni_code = 35282
                if uni_code == 11977:  # 贝
                    uni_code = 36125
                if uni_code == 11979:  # 车
                    uni_code = 26710
                if uni_code == 11987:  # 长
                    uni_code = 38271
                if uni_code == 11988:  # 门
                    uni_code = 38376
                if uni_code == 11992:  # 青
                    uni_code = 38738
                if uni_code == 11993:  # 韦
                    uni_code = 38886
                if uni_code == 11994:  # 页
                    uni_code = 39029
                if uni_code == 11995:  # 风
                    uni_code = 39118
                if uni_code == 11996:  # 飞
                    uni_code = 39134
                if uni_code == 11997:  # 食
                    uni_code = 39135
                if uni_code == 12002:  # 马
                    uni_code = 39532
                if uni_code == 12003:  # 骨
                    uni_code = 39592
                if uni_code == 12004:  # 鬼
                    uni_code = 39740
                if uni_code == 12005:  # 鱼
                    uni_code = 40060
                if uni_code == 12006:  # 鸟
                    uni_code = 40479
                if uni_code == 12007:  # 卤
                    uni_code = 21348
                if uni_code == 12008:  # 麦
                    uni_code = 40614
                if uni_code == 12009:  # 黄
                    uni_code = 40644
                if uni_code == 12012:  # 齐
                    uni_code = 40784
                if uni_code == 12014:  # 齿
                    uni_code = 40831
                if uni_code == 12016:  # 龙
                    uni_code = 40857
                if uni_code == 12019:  # 龟
                    uni_code = 40863
            hanzi_char = chr(uni_code)
            hanzi_chars.append(hanzi_char)
        hanzi_str = ''.join(hanzi_chars)
        return hanzi_str

    def aswhole(self, uni_str):
        uni_str = self.full2half_alphabet_and_num(uni_str)
        uni_str = self.full2half_alphabet_and_num(uni_str)
        uni_str = self.full2half_alphabet_and_num(uni_str)
        uni_str = self.half2full_punctuation(uni_str)
        uni_str = self.kangxi2hanzi(uni_str)
        uni_str = self.buchong2hanzi(uni_str)
        return uni_str


if __name__ == '__main__':
    txtcleaner = TxtCleaner()
    text = txtcleaner.aswhole("第⼀")
    print(text)
